""" A module for general utilities for the package. """

__author__ = "Mihaly Konda"
__version__ = '1.1.2'

# Built-in modules
from collections import UserDict
from collections.abc import Callable
from dataclasses import is_dataclass
from functools import cached_property
import inspect
import os
import sys
from types import FunctionType, MethodType, TracebackType
from typing import Any, Generic, Never, Type, TypeVar

# Qt6 modules
from PySide6.QtCore import Signal


class BijectiveDict(UserDict):
    """
    A custom dictionary providing bijective mapping between a main type
    and a hashable secondary type.
    """

    def __init__(self, primary_type: type) -> None:
        """ Initializer for the class.

        :param primary_type: The type of the main keys for keys() and values().
        """

        super().__init__()
        self._internal_dict = {}
        self._primary_type = primary_type

    def __getitem__(self, item: Any) -> Any:
        """
        Returns a value from the internal dictionary accessed with '[]'
        (either of the main or the secondary type).

        :param item: The key whose associated value is to be returned.

        :returns: The (primary/secondary type) value associated with the key.
        """

        return self._internal_dict[item]

    def __setitem__(self, key: Any, value: Any) -> None:
        """ Sets/updates a key-value pair in a bijective way.

        :param key: Usually the object of primary type.
        :param value: Usually the object of secondary type.
        """

        if key in self._internal_dict:
            dict.__delitem__(self._internal_dict, key)

        if value in self._internal_dict:
            dict.__delitem__(self._internal_dict, value)

        dict.__setitem__(self._internal_dict, key, value)
        dict.__setitem__(self._internal_dict, value, key)

    def __delitem__(self, item: Any) -> None:
        """ Deletes a key-value pair in a bijective way.

        :param item: Usually the object of primary type.
        """

        dict.__delitem__(self._internal_dict, self[item])
        dict.__delitem__(self._internal_dict, item)

    def __len__(self) -> int:
        """ Returns the adjusted size of the internal dictionary. """

        return dict.__len__(self._internal_dict) // 2

    def __repr__(self) -> str:
        """ Returns the repr of the internal dictionary. """

        return dict.__repr__(self._internal_dict)

    def keys(self, primary_only: bool = True) -> list | tuple[list, list]:
        """ Returns the keys of the internal dictionary.

        :param primary_only: A flag to only return the keys of the primary type
            (default behaviour).

        :returns: Either a list of keys of the primary type or additionally
            a list of keys of the secondary type.
        """

        mains = [k for k in self._internal_dict.keys()
                 if isinstance(k, self._primary_type)]

        if primary_only:
            return mains

        secondaries = [k for k in self._internal_dict.keys()
                       if not isinstance(k, self._primary_type)]

        return mains, secondaries

    def values(self, secondary_only: bool = True) -> list | tuple[list, list]:
        """ Returns the values of the internal dictionary.

        :param secondary_only: A flag to only return the keys of the secondary
            type (default behaviour).

        :returns: Either a list of values of the secondary type or additionally
            (as the first item of the return tuple) a list of values of the
            primary type.
        """

        secondaries = [k for k in self._internal_dict.keys()
                       if not isinstance(k, self._primary_type)]

        if secondary_only:
            return secondaries

        mains = [k for k in self._internal_dict.keys()
                 if isinstance(k, self._primary_type)]

        return mains, secondaries


class ReadOnlyDescriptor:
    """ Read-only descriptor for use with protected storage attributes. """

    def __set_name__(self, owner: type, name: str) -> None:
        """ Sets the name of the storage attribute.

        :param owner: The managed class.
        :param name: The managed attribute's name.
        """

        self._storage_name = f"_{name}"

    def __get__(self, instance: Any, instance_type: type = None) -> Any:
        """ Returns the value of the protected storage attribute.

        :param instance: An instance of the managed class (managed instance).
        :param instance_type: Reference to the managed class, unused here.

        :returns: Either the descriptor object itself or the value in the
            storage attribute.
        """

        if instance is None:
            return self  # Class accession: return the descriptor itself

        return getattr(instance, self._storage_name)

    def __set__(self, instance: Any, value: Any) -> Never:
        """
        Sets the value of the protected storage attribute (would, but
        read-only).

        :param instance: An instance of the managed class (managed instance).
        :param value: The value to set for the storage attribute.

        :raises AttributeError: Notifies the user about the attribute being
            read-only.
        """

        raise AttributeError(f"attribute '{self._storage_name[1:]}' of "
                             f"'{instance.__class__.__name__}' object "
                             f"is read-only")


_T = TypeVar('_T')


class SignalBlocker(Generic[_T]):
    """ Temporarily blocks the signals of the handled QObject. """

    def __init__(self, obj: _T) -> None:
        """ Initializer for the class.

        :param obj: An object whose signals should be blocked temporarily.
        """

        self._obj = obj

    def __enter__(self) -> _T:
        """ Blocks the signals of the handled QObject.

        :returns: The handled object.
        """

        self._obj.blockSignals(True)
        return self._obj

    def __exit__(self, exc_type: Type[BaseException] | None,
                 exc_value: BaseException | None,
                 traceback: TracebackType | None) -> None:
        """ Unblocks the signals of the handled QObject. """

        self._obj.blockSignals(False)


class Singleton(type):
    """ A metaclass making instance classes singletons. """

    _instances = {}  # Shared between instance classes, hence the dict

    def __call__(cls, *args, **kwargs) -> Any:
        """ Returns the existing instance or creates a new one. """

        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)

        return cls._instances[cls]


def qt_connect(signal: Callable, slot: Callable, unique: bool = False) \
        -> None:
    """
    Connects a Qt signal to a slot method. It uses the new-style connection
    creation, while having a simple signature and removing the need to sprinkle
    'type: ignore' around connections.

    :param signal: The signal of a Qt-object.
    :param slot: A method of a class serving as a slot for the signal.
    :param unique: Whether the connection should be set to unique. The default
        is False.
    """

    if unique:
        signal.connect(slot, Qt.ConnectionType.UniqueConnection)  # type: ignore
    else:
        signal.connect(slot)  # type: ignore


def resource_path(relative_path: str) -> str:
    """ Get absolute path to resource (for apps built with PyInstaller).

    :param relative_path: A string representing a relative path to the
        requested resource.

    :returns: The absolute path to the requested resource (as a simple string).
    """

    # PyInstaller creates a temp folder and stores path in _MEIPASS
    return os.path.join(getattr(sys, '_MEIPASS', os.path.abspath('')),
                        relative_path)


def _stub_repr_function_like(f: cached_property | FunctionType | MethodType,
                             class_bound: bool) -> str:
    """ Creates a stub representation for a function-like object.

    :param f: The function-like object whose stub representation is to be made.
    :param class_bound: A flag to add 'self' to the representation.

    :returns: The stub representation of the input function-like object.
    """

    decorator = ''
    if isinstance(f, cached_property):
        f = f.func
        decorator = '\t@cached_property\n'
    elif isinstance(f, MethodType):  # Cannot yet differentiate static methods
        decorator = '\t@classmethod\n'

    anns = []
    defaults = [] if f.__defaults__ is None else list(f.__defaults__)

    # Using reverse order to get the default values
    for param, typ in list(f.__annotations__.items())[::-1]:
        if param == 'return':  # Return annotation is handled on its own
            continue

        # The type is passed as a string if the source has __future__ annot.
        if not isinstance(typ, str) and typ is not None:
            typ = typ.__name__

        default_value = ''
        if defaults:
            dv = defaults.pop()
            if isinstance(dv, str):
                default_value = f" = '{dv}'"
            else:
                default_value = f" = {dv}"

        anns.append(f"{param}: {typ}{default_value}")

    if isinstance(f, MethodType):  # Refers to a class method
        anns.append('cls')
    elif class_bound:
        anns.append('self')

    try:
        ret_type = f.__annotations__['return']
    except KeyError:  # Return annotation is not provided in the function def
        return_annotation = ''
    else:
        # The type is passed as a string if the source has __future__ annot.
        if not isinstance(ret_type, str) and ret_type is not None:
            ret_type = ret_type.__name__
        return_annotation = f" -> {ret_type}"

    func_repr = f"{decorator}{'\t' if class_bound else ''}" \
                f"def {f.__name__}({', '.join(anns[::-1])}" \
                f"){return_annotation}: ...\n"

    return func_repr


def stub_repr(obj: object, signals: list[str] | None = None,
              extra_cvs: str | None = None, add_getattr: bool = False) -> str:
    """ Creates a specifically formatted stub representation of an object.

    .. note:: Static methods must be changed to class methods for them to be
        compatible with the function (which cannot yet handle static methods).

    :param obj: The object whose stub representation is to be made.
    :param signals: A list of strings representing the (Qt) signals of the
        object, as [sigName(carriedType1, ...)]. The default is None.
    :param extra_cvs: A string of extra class variables that are not defined
        as CVs in the source code. The default is None.
    :param add_getattr: A flag for adding a special __getattr__ method to the
        repr. For public Qt-based classes, it needs so that the static type
        checker does not warn about variables declared outside __init__.
        The default is False.

    :returns: The stub representation of the input object.
    """

    repr_ = ''
    if inspect.isclass(obj):
        function_likes = []
        class_vars = []
        signal_reprs = []
        properties = []
        for dir_item in dir(obj):
            if not dir_item.startswith('__'):
                cls_attr = getattr(obj, dir_item)
                # print(f"{str(obj):<40}{str(cls_attr):<75}{type(cls_attr)}")
                if any(isinstance(cls_attr, typ) for typ in
                       [cached_property, FunctionType, MethodType]):
                    function_likes.append(
                        _stub_repr_function_like(cls_attr, True))
                elif isinstance(cls_attr, ReadOnlyDescriptor):
                    repr_ = f"\t{dir_item}: ReadOnlyDescriptor = " \
                            "ReadOnlyDescriptor()\n"
                    class_vars.append(repr_)
                elif isinstance(cls_attr, Signal) and signals is not None:
                    for sig in signals:
                        if (sig_name := sig.split('(')[0]) in str(cls_attr):
                            signal_reprs.append(f"\t{sig_name} : "
                                                "ClassVar[Signal] = "
                                                f"...  # {sig}\n")
                elif isinstance(cls_attr, property):
                    prop_funcs = {'fget': 'property',
                                  'fset': '{fn}.setter',
                                  'fdel': '{fn}.deleter'}

                    for attr, deco in prop_funcs.items():
                        if (func := getattr(cls_attr, attr)) is not None:
                            if '{' in deco:
                                deco = deco.format(fn=func.__name__)

                            repr_ = f"\t@{deco}\n" \
                                    f"{_stub_repr_function_like(func, True)}"
                            properties.append(repr_)

        bases = ''
        if obj.__bases__[0].__name__ != 'object':
            bases = ', '.join([b.__name__ for b in obj.__bases__])
            if isinstance(obj, Singleton):
                bases += f", metaclass={type(obj).__name__}"

            bases = f'({bases})'
        elif isinstance(obj, Singleton):
            bases = f'(metaclass={type(obj).__name__})'

        class_decorator = '@dataclass\n' if is_dataclass(obj) else ''

        extra_getattr = ''
        if add_getattr:
            extra_getattr = "\tdef __getattr__(self, name: str) -> Any: ...\n"

        repr_ = f"{class_decorator}class {obj.__name__}{bases}:\n" \
                f"{''.join(signal_reprs) + '\n' if signal_reprs else ''}" \
                f"{extra_cvs + '\n' if extra_cvs is not None else ''}" \
                f"{''.join(class_vars) + '\n' if class_vars else ''}" \
                f"{_stub_repr_function_like(obj.__init__, True)}" \
                f"{extra_getattr}" \
                f"{''.join(function_likes)}" \
                f"{''.join(properties)}"
    elif inspect.isfunction(obj):
        repr_ = _stub_repr_function_like(obj, False)

    return repr_



from dataclasses import dataclass
from functools import cached_property
from typing import ClassVar, Optional
from PySide6.QtCore import Signal, Qt
from PySide6.QtGui import QColor, QIcon, QKeyEvent, QMouseEvent, QPaintEvent
from PySide6.QtWidgets import QDialog, QDockWidget, QMainWindow, QWidget
from src.utils_qt_mk import ReadOnlyDescriptor, Singleton
from src.utils_qt_mk import ThemeParameters


Colours: _Colours = None


def text_colour_threshold() -> int: ...
def set_text_colour_threshold(new_value: int) -> None: ...
def icon_file_path() -> str: ...
def set_icon_file_path(new_path: str = '') -> None: ...
def extended_default() -> bool: ...
def set_extended_default(new_default: bool) -> None: ...


class Colour:
	b: ReadOnlyDescriptor = ReadOnlyDescriptor()
	g: ReadOnlyDescriptor = ReadOnlyDescriptor()
	name: ReadOnlyDescriptor = ReadOnlyDescriptor()
	r: ReadOnlyDescriptor = ReadOnlyDescriptor()

	def __init__(self, name: str = 'white', r: int = 255, g: int = 255, b: int = 255) -> None: ...
	@cached_property
	def as_hex(self) -> str: ...
	def as_qt(self, negative: bool = False) -> QColor: ...
	@cached_property
	def as_rgb(self) -> str: ...
	def colour_box(self, width: int = 20, height: int = 20) -> QIcon: ...
	def text_colour(self) -> Qt.GlobalColor: ...


class _Colours(metaclass=Singleton):
	aliceblue: Colour = None
	antiquewhite: Colour = None
	antiquewhite1: Colour = None
	antiquewhite2: Colour = None
	antiquewhite3: Colour = None
	antiquewhite4: Colour = None
	aquamarine: Colour = None
	aquamarine1: Colour = None
	aquamarine2: Colour = None
	aquamarine3: Colour = None
	aquamarine4: Colour = None
	azure: Colour = None
	azure1: Colour = None
	azure2: Colour = None
	azure3: Colour = None
	azure4: Colour = None
	beige: Colour = None
	bisque: Colour = None
	bisque1: Colour = None
	bisque2: Colour = None
	bisque3: Colour = None
	bisque4: Colour = None
	black: Colour = None
	blanchedalmond: Colour = None
	blue: Colour = None
	blue1: Colour = None
	blue2: Colour = None
	blue3: Colour = None
	blue4: Colour = None
	blueviolet: Colour = None
	brown: Colour = None
	brown1: Colour = None
	brown2: Colour = None
	brown3: Colour = None
	brown4: Colour = None
	burlywood: Colour = None
	burlywood1: Colour = None
	burlywood2: Colour = None
	burlywood3: Colour = None
	burlywood4: Colour = None
	cadetblue: Colour = None
	cadetblue1: Colour = None
	cadetblue2: Colour = None
	cadetblue3: Colour = None
	cadetblue4: Colour = None
	chartreuse: Colour = None
	chartreuse1: Colour = None
	chartreuse2: Colour = None
	chartreuse3: Colour = None
	chartreuse4: Colour = None
	chocolate: Colour = None
	chocolate1: Colour = None
	chocolate2: Colour = None
	chocolate3: Colour = None
	chocolate4: Colour = None
	coral: Colour = None
	coral1: Colour = None
	coral2: Colour = None
	coral3: Colour = None
	coral4: Colour = None
	cornflowerblue: Colour = None
	cornsilk: Colour = None
	cornsilk1: Colour = None
	cornsilk2: Colour = None
	cornsilk3: Colour = None
	cornsilk4: Colour = None
	cyan: Colour = None
	cyan1: Colour = None
	cyan2: Colour = None
	cyan3: Colour = None
	cyan4: Colour = None
	darkblue: Colour = None
	darkcyan: Colour = None
	darkgoldenrod: Colour = None
	darkgoldenrod1: Colour = None
	darkgoldenrod2: Colour = None
	darkgoldenrod3: Colour = None
	darkgoldenrod4: Colour = None
	darkgreen: Colour = None
	darkgrey: Colour = None
	darkkhaki: Colour = None
	darkmagenta: Colour = None
	darkolivegreen: Colour = None
	darkolivegreen1: Colour = None
	darkolivegreen2: Colour = None
	darkolivegreen3: Colour = None
	darkolivegreen4: Colour = None
	darkorange: Colour = None
	darkorange1: Colour = None
	darkorange2: Colour = None
	darkorange3: Colour = None
	darkorange4: Colour = None
	darkorchid: Colour = None
	darkorchid1: Colour = None
	darkorchid2: Colour = None
	darkorchid3: Colour = None
	darkorchid4: Colour = None
	darkred: Colour = None
	darksalmon: Colour = None
	darkseagreen: Colour = None
	darkseagreen1: Colour = None
	darkseagreen2: Colour = None
	darkseagreen3: Colour = None
	darkseagreen4: Colour = None
	darkslateblue: Colour = None
	darkslategrey: Colour = None
	darkslategrey1: Colour = None
	darkslategrey2: Colour = None
	darkslategrey3: Colour = None
	darkslategrey4: Colour = None
	darkturquoise: Colour = None
	darkviolet: Colour = None
	deeppink: Colour = None
	deeppink1: Colour = None
	deeppink2: Colour = None
	deeppink3: Colour = None
	deeppink4: Colour = None
	deepskyblue: Colour = None
	deepskyblue1: Colour = None
	deepskyblue2: Colour = None
	deepskyblue3: Colour = None
	deepskyblue4: Colour = None
	dimgrey: Colour = None
	dodgerblue: Colour = None
	dodgerblue1: Colour = None
	dodgerblue2: Colour = None
	dodgerblue3: Colour = None
	dodgerblue4: Colour = None
	firebrick: Colour = None
	firebrick1: Colour = None
	firebrick2: Colour = None
	firebrick3: Colour = None
	firebrick4: Colour = None
	floralwhite: Colour = None
	forestgreen: Colour = None
	gainsboro: Colour = None
	ghostwhite: Colour = None
	gold: Colour = None
	gold1: Colour = None
	gold2: Colour = None
	gold3: Colour = None
	gold4: Colour = None
	goldenrod: Colour = None
	goldenrod1: Colour = None
	goldenrod2: Colour = None
	goldenrod3: Colour = None
	goldenrod4: Colour = None
	grey: Colour = None
	grey0: Colour = None
	grey1: Colour = None
	grey2: Colour = None
	grey3: Colour = None
	grey4: Colour = None
	grey5: Colour = None
	grey6: Colour = None
	grey7: Colour = None
	grey8: Colour = None
	grey9: Colour = None
	grey10: Colour = None
	grey11: Colour = None
	grey12: Colour = None
	grey13: Colour = None
	grey14: Colour = None
	grey15: Colour = None
	grey16: Colour = None
	grey17: Colour = None
	grey18: Colour = None
	grey19: Colour = None
	grey20: Colour = None
	grey21: Colour = None
	grey22: Colour = None
	grey23: Colour = None
	grey24: Colour = None
	grey25: Colour = None
	grey26: Colour = None
	grey27: Colour = None
	grey28: Colour = None
	grey29: Colour = None
	grey30: Colour = None
	grey31: Colour = None
	grey32: Colour = None
	grey33: Colour = None
	grey34: Colour = None
	grey35: Colour = None
	grey36: Colour = None
	grey37: Colour = None
	grey38: Colour = None
	grey39: Colour = None
	grey40: Colour = None
	grey41: Colour = None
	grey42: Colour = None
	grey43: Colour = None
	grey44: Colour = None
	grey45: Colour = None
	grey46: Colour = None
	grey47: Colour = None
	grey48: Colour = None
	grey49: Colour = None
	grey50: Colour = None
	grey51: Colour = None
	grey52: Colour = None
	grey53: Colour = None
	grey54: Colour = None
	grey55: Colour = None
	grey56: Colour = None
	grey57: Colour = None
	grey58: Colour = None
	grey59: Colour = None
	grey60: Colour = None
	grey61: Colour = None
	grey62: Colour = None
	grey63: Colour = None
	grey64: Colour = None
	grey65: Colour = None
	grey66: Colour = None
	grey67: Colour = None
	grey68: Colour = None
	grey69: Colour = None
	grey70: Colour = None
	grey71: Colour = None
	grey72: Colour = None
	grey73: Colour = None
	grey74: Colour = None
	grey75: Colour = None
	grey76: Colour = None
	grey77: Colour = None
	grey78: Colour = None
	grey79: Colour = None
	grey80: Colour = None
	grey81: Colour = None
	grey82: Colour = None
	grey83: Colour = None
	grey84: Colour = None
	grey85: Colour = None
	grey86: Colour = None
	grey87: Colour = None
	grey88: Colour = None
	grey89: Colour = None
	grey90: Colour = None
	grey91: Colour = None
	grey92: Colour = None
	grey93: Colour = None
	grey94: Colour = None
	grey95: Colour = None
	grey96: Colour = None
	grey97: Colour = None
	grey98: Colour = None
	grey99: Colour = None
	grey100: Colour = None
	green: Colour = None
	green1: Colour = None
	green2: Colour = None
	green3: Colour = None
	green4: Colour = None
	greenyellow: Colour = None
	honeydew: Colour = None
	honeydew1: Colour = None
	honeydew2: Colour = None
	honeydew3: Colour = None
	honeydew4: Colour = None
	hotpink: Colour = None
	hotpink1: Colour = None
	hotpink2: Colour = None
	hotpink3: Colour = None
	hotpink4: Colour = None
	indianred: Colour = None
	indianred1: Colour = None
	indianred2: Colour = None
	indianred3: Colour = None
	indianred4: Colour = None
	ivory: Colour = None
	ivory1: Colour = None
	ivory2: Colour = None
	ivory3: Colour = None
	ivory4: Colour = None
	khaki: Colour = None
	khaki1: Colour = None
	khaki2: Colour = None
	khaki3: Colour = None
	khaki4: Colour = None
	lavender: Colour = None
	lavenderblush: Colour = None
	lavenderblush1: Colour = None
	lavenderblush2: Colour = None
	lavenderblush3: Colour = None
	lavenderblush4: Colour = None
	lawngreen: Colour = None
	lemonchiffon: Colour = None
	lemonchiffon1: Colour = None
	lemonchiffon2: Colour = None
	lemonchiffon3: Colour = None
	lemonchiffon4: Colour = None
	lightblue: Colour = None
	lightblue1: Colour = None
	lightblue2: Colour = None
	lightblue3: Colour = None
	lightblue4: Colour = None
	lightcoral: Colour = None
	lightcyan: Colour = None
	lightcyan1: Colour = None
	lightcyan2: Colour = None
	lightcyan3: Colour = None
	lightcyan4: Colour = None
	lightgoldenrod: Colour = None
	lightgoldenrod1: Colour = None
	lightgoldenrod2: Colour = None
	lightgoldenrod3: Colour = None
	lightgoldenrod4: Colour = None
	lightgoldenrodyellow: Colour = None
	lightgreen: Colour = None
	lightgrey: Colour = None
	lightpink: Colour = None
	lightpink1: Colour = None
	lightpink2: Colour = None
	lightpink3: Colour = None
	lightpink4: Colour = None
	lightsalmon: Colour = None
	lightsalmon1: Colour = None
	lightsalmon2: Colour = None
	lightsalmon3: Colour = None
	lightsalmon4: Colour = None
	lightseagreen: Colour = None
	lightskyblue: Colour = None
	lightskyblue1: Colour = None
	lightskyblue2: Colour = None
	lightskyblue3: Colour = None
	lightskyblue4: Colour = None
	lightslateblue: Colour = None
	lightslategrey: Colour = None
	lightsteelblue: Colour = None
	lightsteelblue1: Colour = None
	lightsteelblue2: Colour = None
	lightsteelblue3: Colour = None
	lightsteelblue4: Colour = None
	lightyellow: Colour = None
	lightyellow1: Colour = None
	lightyellow2: Colour = None
	lightyellow3: Colour = None
	lightyellow4: Colour = None
	limegreen: Colour = None
	linen: Colour = None
	magenta: Colour = None
	magenta1: Colour = None
	magenta2: Colour = None
	magenta3: Colour = None
	magenta4: Colour = None
	maroon: Colour = None
	maroon1: Colour = None
	maroon2: Colour = None
	maroon3: Colour = None
	maroon4: Colour = None
	mediumaquamarine: Colour = None
	mediumblue: Colour = None
	mediumorchid: Colour = None
	mediumorchid1: Colour = None
	mediumorchid2: Colour = None
	mediumorchid3: Colour = None
	mediumorchid4: Colour = None
	mediumpurple: Colour = None
	mediumpurple1: Colour = None
	mediumpurple2: Colour = None
	mediumpurple3: Colour = None
	mediumpurple4: Colour = None
	mediumseagreen: Colour = None
	mediumslateblue: Colour = None
	mediumspringgreen: Colour = None
	mediumturquoise: Colour = None
	mediumvioletred: Colour = None
	midnightblue: Colour = None
	mintcream: Colour = None
	mistyrose: Colour = None
	mistyrose1: Colour = None
	mistyrose2: Colour = None
	mistyrose3: Colour = None
	mistyrose4: Colour = None
	moccasin: Colour = None
	navajowhite: Colour = None
	navajowhite1: Colour = None
	navajowhite2: Colour = None
	navajowhite3: Colour = None
	navajowhite4: Colour = None
	navy: Colour = None
	navyblue: Colour = None
	oldlace: Colour = None
	olivedrab: Colour = None
	olivedrab1: Colour = None
	olivedrab2: Colour = None
	olivedrab3: Colour = None
	olivedrab4: Colour = None
	orange: Colour = None
	orange1: Colour = None
	orange2: Colour = None
	orange3: Colour = None
	orange4: Colour = None
	orangered: Colour = None
	orangered1: Colour = None
	orangered2: Colour = None
	orangered3: Colour = None
	orangered4: Colour = None
	orchid: Colour = None
	orchid1: Colour = None
	orchid2: Colour = None
	orchid3: Colour = None
	orchid4: Colour = None
	palegoldenrod: Colour = None
	palegreen: Colour = None
	palegreen1: Colour = None
	palegreen2: Colour = None
	palegreen3: Colour = None
	palegreen4: Colour = None
	paleturquoise: Colour = None
	paleturquoise1: Colour = None
	paleturquoise2: Colour = None
	paleturquoise3: Colour = None
	paleturquoise4: Colour = None
	palevioletred: Colour = None
	palevioletred1: Colour = None
	palevioletred2: Colour = None
	palevioletred3: Colour = None
	palevioletred4: Colour = None
	papayawhip: Colour = None
	peachpuff: Colour = None
	peachpuff1: Colour = None
	peachpuff2: Colour = None
	peachpuff3: Colour = None
	peachpuff4: Colour = None
	peru: Colour = None
	pink: Colour = None
	pink1: Colour = None
	pink2: Colour = None
	pink3: Colour = None
	pink4: Colour = None
	plum: Colour = None
	plum1: Colour = None
	plum2: Colour = None
	plum3: Colour = None
	plum4: Colour = None
	powderblue: Colour = None
	purple: Colour = None
	purple1: Colour = None
	purple2: Colour = None
	purple3: Colour = None
	purple4: Colour = None
	red: Colour = None
	red1: Colour = None
	red2: Colour = None
	red3: Colour = None
	red4: Colour = None
	rosybrown: Colour = None
	rosybrown1: Colour = None
	rosybrown2: Colour = None
	rosybrown3: Colour = None
	rosybrown4: Colour = None
	royalblue: Colour = None
	royalblue1: Colour = None
	royalblue2: Colour = None
	royalblue3: Colour = None
	royalblue4: Colour = None
	saddlebrown: Colour = None
	salmon: Colour = None
	salmon1: Colour = None
	salmon2: Colour = None
	salmon3: Colour = None
	salmon4: Colour = None
	sandybrown: Colour = None
	seagreen: Colour = None
	seagreen1: Colour = None
	seagreen2: Colour = None
	seagreen3: Colour = None
	seagreen4: Colour = None
	seashell: Colour = None
	seashell1: Colour = None
	seashell2: Colour = None
	seashell3: Colour = None
	seashell4: Colour = None
	sienna: Colour = None
	sienna1: Colour = None
	sienna2: Colour = None
	sienna3: Colour = None
	sienna4: Colour = None
	skyblue: Colour = None
	skyblue1: Colour = None
	skyblue2: Colour = None
	skyblue3: Colour = None
	skyblue4: Colour = None
	slateblue: Colour = None
	slateblue1: Colour = None
	slateblue2: Colour = None
	slateblue3: Colour = None
	slateblue4: Colour = None
	slategrey: Colour = None
	slategrey1: Colour = None
	slategrey2: Colour = None
	slategrey3: Colour = None
	slategrey4: Colour = None
	snow: Colour = None
	snow1: Colour = None
	snow2: Colour = None
	snow3: Colour = None
	snow4: Colour = None
	springgreen: Colour = None
	springgreen1: Colour = None
	springgreen2: Colour = None
	springgreen3: Colour = None
	springgreen4: Colour = None
	steelblue: Colour = None
	steelblue1: Colour = None
	steelblue2: Colour = None
	steelblue3: Colour = None
	steelblue4: Colour = None
	tan: Colour = None
	tan1: Colour = None
	tan2: Colour = None
	tan3: Colour = None
	tan4: Colour = None
	thistle: Colour = None
	thistle1: Colour = None
	thistle2: Colour = None
	thistle3: Colour = None
	thistle4: Colour = None
	tomato: Colour = None
	tomato1: Colour = None
	tomato2: Colour = None
	tomato3: Colour = None
	tomato4: Colour = None
	turquoise: Colour = None
	turquoise1: Colour = None
	turquoise2: Colour = None
	turquoise3: Colour = None
	turquoise4: Colour = None
	violet: Colour = None
	violetred: Colour = None
	violetred1: Colour = None
	violetred2: Colour = None
	violetred3: Colour = None
	violetred4: Colour = None
	wheat: Colour = None
	wheat1: Colour = None
	wheat2: Colour = None
	wheat3: Colour = None
	wheat4: Colour = None
	white: Colour = None
	whitesmoke: Colour = None
	yellow: Colour = None
	yellow1: Colour = None
	yellow2: Colour = None
	yellow3: Colour = None
	yellow4: Colour = None
	yellowgreen: Colour = None
	def __init__(self) -> None: ...
	def colour_at(self, idx: int) -> Colour: ...
	def from_qt(self, qc: QColor) -> Colour: ...
	def index(self, name: str) -> int: ...


@dataclass
class _ColourBoxData:
	def __init__(self, row: int = -1, column: int = -1, colour: Optional[Colour] = None) -> None: ...


class _ColourBoxDrawer(QWidget):
	colourSelected : ClassVar[Signal] = ...  # colourSelected(int)

	def __init__(self, default_colour: Colour) -> None: ...
	def keyPressEvent(self, event: QKeyEvent) -> None: ...
	def mousePressEvent(self, event: QMouseEvent) -> None: ...
	def paintEvent(self, event: QPaintEvent) -> None: ...
	@property
	def selection(self) -> Colour: ...
	@selection.setter
	def selection(self, new_selection: _ColourBoxData) -> None: ...


class _ColourSelectorMixin:
	colourChanged : ClassVar[Signal] = ...  # colourChanged(int, Colour)

	def __init__(self, button_id: int = 0, default_colour: Colour = Colour('white', 255, 255, 255), widget_theme: ThemeParameters = None) -> None: ...
	def _setup_connections(self) -> None: ...
	def _setup_ui(self) -> None: ...
	def _slot_apply(self) -> None: ...
	def _slot_cancel(self) -> None: ...
	def _slot_filter(self) -> None: ...
	def _slot_tab_changed(self, index: int) -> None: ...
	def _slot_update_selection(self, index: int) -> None: ...
	@property
	def theme(self) -> ThemeParameters: ...
	@theme.setter
	def theme(self, new_theme: ThemeParameters) -> None: ...


class ColourSelector(_ColourSelectorMixin, QDialog):
	def __init__(self, button_id: int = 0, default_colour: Colour = Colour('white', 255, 255, 255), widget_theme: ThemeParameters = None) -> None: ...
	def _setup_connections(self) -> None: ...
	def _setup_ui(self) -> None: ...
	def _slot_apply(self) -> None: ...
	def _slot_cancel(self) -> None: ...
	def _slot_filter(self) -> None: ...
	def _slot_tab_changed(self, index: int) -> None: ...
	def _slot_update_selection(self, index: int) -> None: ...
	@property
	def theme(self) -> ThemeParameters: ...
	@theme.setter
	def theme(self, new_theme: ThemeParameters) -> None: ...


class ColourSelectorDW(_ColourSelectorMixin, QDockWidget):
	def __init__(self, button_id: int = 0, default_colour: Colour = Colour('white', 255, 255, 255), widget_theme: ThemeParameters = None) -> None: ...
	def _setup_connections(self) -> None: ...
	def _setup_ui(self) -> None: ...
	def _slot_apply(self) -> None: ...
	def _slot_cancel(self) -> None: ...
	def _slot_filter(self) -> None: ...
	def _slot_tab_changed(self, index: int) -> None: ...
	def _slot_update_selection(self, index: int) -> None: ...
	@property
	def theme(self) -> ThemeParameters: ...
	@theme.setter
	def theme(self, new_theme: ThemeParameters) -> None: ...


class _ColourScale(QWidget):
	def __init__(self, colours: list[Colour] = None, steps: int = 0, horizontal: bool = False) -> None: ...
	@classmethod
	def _segment_calculator(cls, colours: tuple[Colour], steps: int) -> list[QColor]: ...
	def paintEvent(self, event: QPaintEvent) -> None: ...
	def update_scale(self, colours: list[Colour], steps: int) -> None: ...


class _ColourScaleCreatorMixin:
	colourScaleChanged : ClassVar[Signal] = ...  # colourScaleChanged(list)

	def __init__(self, colours: list[Colour] = None, horizontal: bool = False, widget_theme: ThemeParameters = None, parent: QMainWindow = None) -> None: ...
	def _setup_connections(self) -> None: ...
	def _setup_ui(self) -> None: ...
	def _slot_add_colour(self) -> None: ...
	def _slot_apply(self) -> None: ...
	def _slot_cancel(self) -> None: ...
	def _slot_export_scale(self) -> None: ...
	def _slot_import_scale(self) -> None: ...
	def _slot_remove_colour(self) -> None: ...
	def _slot_update_scale(self) -> None: ...
	def _slot_update_total_steps(self) -> None: ...
	@property
	def theme(self) -> ThemeParameters: ...
	@theme.setter
	def theme(self, new_theme: ThemeParameters) -> None: ...


class ColourScaleCreator(_ColourScaleCreatorMixin, QDialog):
	def __init__(self, colours: list[Colour] = None, horizontal: bool = False, widget_theme: ThemeParameters = None) -> None: ...
	def _setup_connections(self) -> None: ...
	def _setup_ui(self) -> None: ...
	def _slot_add_colour(self) -> None: ...
	def _slot_apply(self) -> None: ...
	def _slot_cancel(self) -> None: ...
	def _slot_export_scale(self) -> None: ...
	def _slot_import_scale(self) -> None: ...
	def _slot_remove_colour(self) -> None: ...
	def _slot_update_scale(self) -> None: ...
	def _slot_update_total_steps(self) -> None: ...
	@property
	def theme(self) -> ThemeParameters: ...
	@theme.setter
	def theme(self, new_theme: ThemeParameters) -> None: ...


class ColourScaleCreatorDW(_ColourScaleCreatorMixin, QDockWidget):
	def __init__(self, parent: QMainWindow, colours: list[Colour] = None, horizontal: bool = False, widget_theme: ThemeParameters = None) -> None: ...
	def _setup_connections(self) -> None: ...
	def _setup_ui(self) -> None: ...
	def _slot_add_colour(self) -> None: ...
	def _slot_apply(self) -> None: ...
	def _slot_cancel(self) -> None: ...
	def _slot_export_scale(self) -> None: ...
	def _slot_import_scale(self) -> None: ...
	def _slot_remove_colour(self) -> None: ...
	def _slot_update_scale(self) -> None: ...
	def _slot_update_total_steps(self) -> None: ...
	@property
	def theme(self) -> ThemeParameters: ...
	@theme.setter
	def theme(self, new_theme: ThemeParameters) -> None: ...


class _TestApplication(QMainWindow):
	def __init__(self) -> None: ...
	def _setup_connections(self) -> None: ...
	def _setup_ui(self) -> None: ...
	def _slot_cs_test(self) -> None: ...
	def _slot_csc_test(self) -> None: ...



from dataclasses import dataclass
from typing import TypeVar
from PySide6.QtWidgets import QWidget
from utils_qt_mk._general import Singleton


WidgetTheme: _WidgetTheme = None
QWidgetT = TypeVar('QWidgetT', bound=QWidget)


def set_widget_theme(widget: QWidgetT, theme: ThemeParameters = None) -> None: ...


@dataclass
class ThemeParameters:
	def __init__(self, src_file: str | None = None) -> None: ...
	def write_json(self, destination: str) -> None: ...


class _WidgetTheme(metaclass=Singleton):
	dark: ThemeParameters = None
	light: ThemeParameters = None
	matrix: ThemeParameters = None
	yellow: ThemeParameters = None
	def __init__(self) -> None: ...
	def load_dict(self) -> None: ...



""" A module for a theme creator dialog. """

__author__ = "Mihaly Konda"
__version__ = '1.0.4'

# Built-in modules
from dataclasses import fields
import os
import sys

# Qt6 modules
from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

# Custom modules
from utils_qt_mk.config import _PACKAGE_DIR, theme_dir
_THEME_DIR = theme_dir()

from utils_qt_mk.colours import ColourSelector, set_extended_default
from utils_qt_mk.custom_file_dialog import custom_dialog, CFDType
from utils_qt_mk.general import SignalBlocker, stub_repr, qt_connect
from utils_qt_mk.theme import set_widget_theme, ThemeParameters, WidgetTheme


set_extended_default(True)  # To show the extended selector by default


class _ColourSetter(QWidget):
    """ A widget for colour selection or manual colour setting. """

    def __init__(self) -> None:
        """ Initializer for the class. """

        super().__init__(parent=None)

        self._set_colour = QColor(255, 255, 255)  # type: ignore

        self._setup_ui()
        self._setup_connections()

    def _setup_ui(self) -> None:
        """ Sets up the user interface: GUI objects and layouts. """

        # GUI objects
        self._chkSelector = QCheckBox(text="Use selector", parent=None)
        self._chkSelector.setChecked(True)
        self._lblColour = QLabel(parent=None)
        self._set_colour_label()

        self._btnSelector = QPushButton("Open selector dialog")
        self._btnSelector.setFixedHeight(25)
        self._btnSelector.setObjectName('button')
        self._lblRGB = QLabel(text='RGB', parent=None)
        self._spblistRGB = [QSpinBox() for _ in range(3)]
        for spb in self._spblistRGB:
            spb.setRange(0, 255)
            spb.setValue(255)
            spb.setObjectName('spinbox')

        # Layouts
        self._vloSelector = QVBoxLayout()  # So the button is in line with...
        self._vloSelector.addWidget(self._btnSelector)  # ... everything else
        self._vloSelector.setContentsMargins(0, 0, 0, 0)
        self._vloSelector.setSpacing(0)

        self._wdgSelector = QWidget()
        self._wdgSelector.setLayout(self._vloSelector)

        self._hloCustomColour = QHBoxLayout()
        self._hloCustomColour.addWidget(self._lblRGB)
        for spb in self._spblistRGB:
            self._hloCustomColour.addWidget(spb)

        self._hloCustomColour.setContentsMargins(0, 0, 0, 0)
        self._hloCustomColour.setSpacing(0)

        self._wdgCustomColour = QWidget()
        self._wdgCustomColour.setLayout(self._hloCustomColour)

        self._sloStackedLayout = QStackedLayout()
        self._sloStackedLayout.addWidget(self._wdgSelector)
        self._sloStackedLayout.addWidget(self._wdgCustomColour)

        self._hloMainLayout = QHBoxLayout()
        self._hloMainLayout.addWidget(self._chkSelector)
        self._hloMainLayout.addWidget(self._lblColour)
        self._hloMainLayout.addLayout(self._sloStackedLayout)

        self.setLayout(self._hloMainLayout)

    def _setup_connections(self) -> None:
        """ Sets up the connections of the GUI objects. """

        qt_connect(self._chkSelector.stateChanged, self._slot_update_selector)
        qt_connect(self._btnSelector.clicked, self._slot_colour_selector)
        for spb in self._spblistRGB:
            qt_connect(spb.valueChanged, self._update_colour)

    @property
    def colour(self) -> QColor:
        """ Returns the stored colour. """

        return self._set_colour

    @colour.setter
    def colour(self, new_colour: QColor) -> None:
        """ Updates the stored colour and its associated widgets.

        :param new_colour: The new colour to set.
        """

        self._set_colour = new_colour
        self._update_spinboxes()
        self._set_colour_label()

    def set_enabled(self, new_state: bool) -> None:
        """ Sets the enabled state of the controls.

        :param new_state: The new enabled state to set.
        """

        self._chkSelector.setEnabled(new_state)
        self._btnSelector.setEnabled(new_state)
        for spb in self._spblistRGB:
            spb.setEnabled(new_state)

    def _slot_update_selector(self) -> None:
        """ Updates which selector is shown based on the control checkbox. """

        if self._chkSelector.isChecked():
            self._sloStackedLayout.setCurrentIndex(0)
        else:
            self._sloStackedLayout.setCurrentIndex(1)

    def _set_colour_label(self) -> None:
        """ Sets the pixmap of the colour's display label. """

        pixmap = QPixmap(20, 20)  # type: ignore
        pixmap.fill(self._set_colour)
        self._lblColour.setPixmap(pixmap)

    def _slot_colour_selector(self) -> None:
        """ Shows a colour selector dialog. """

        def catch_signal(button_id, colour) -> None:
            """ Catches the signal carrying the newly set colour.

            :param button_id: The caller button's ID, unused here.
            :param colour: The colour to set.
            """

            self._set_colour = colour.as_qt()
            self._update_colour()

        cs = ColourSelector()
        cs.colourChanged.connect(catch_signal)
        cs.exec()

    def _update_spinboxes(self) -> None:
        """ Updates the values of spinboxes by the stored colour. """

        channels = {0: 'red', 1: 'green', 2: 'blue'}
        for idx, spb in enumerate(self._spblistRGB):
            with SignalBlocker(spb) as obj:
                obj.setValue(getattr(self._set_colour, channels[idx])())

    def _update_colour(self) -> None:
        """
        Updates the stored colour and its display label according to the sender.
        """

        if self.sender().objectName() == 'button':
            # Colour set in nested catch_signal()
            self._update_spinboxes()
        elif self.sender().objectName() == 'spinbox':
            self._set_colour = QColor(
                *[s.value() for s in self._spblistRGB])  # type: ignore

        self._set_colour_label()


class _ThemePreview(QWidget):
    """ A widget for previewing a theme. """

    def __init__(self) -> None:
        """ Initializer for the class. """

        super().__init__(parent=None)

        self._setup_ui()

    def _setup_ui(self) -> None:
        """ Sets up the user interface: GUI objects and layouts. """

        # GUI objects
        self._lblTest = QLabel(text="Test label (with tooltip)", parent=None)
        self._lblTest.setToolTip("Test tooltip")
        self._cmbTest = QComboBox(parent=None)
        self._cmbTest.addItems(["Item 1", "Item 2", "Item 3"])
        self._chkTest = QCheckBox(text="Test checkbox", parent=None)
        self._chkTest.setTristate(True)
        self._ledTest = QLineEdit()
        self._ledTest.setPlaceholderText('Placeholder')
        self._ledTest2 = QLineEdit()
        self._ledTest2.setText("Test text")
        self._ledTest2.setSelection(5, 4)
        self._btnTest = QPushButton("Test button")
        self._sldTest = QSlider(Qt.Orientation.Horizontal, parent=None)
        self._sldTest.setValue(50)
        self._pbTest = QProgressBar(parent=None)
        self._pbTest.setValue(35)
        self._pbTest2 = QProgressBar(parent=None)
        self._pbTest2.setValue(85)

        # Layouts
        self._vloMainLayout = QVBoxLayout()
        self._vloMainLayout.addWidget(self._lblTest)
        self._vloMainLayout.addWidget(self._cmbTest)
        self._vloMainLayout.addWidget(self._chkTest)
        self._vloMainLayout.addWidget(self._ledTest)
        self._vloMainLayout.addWidget(self._ledTest2)
        self._vloMainLayout.addWidget(self._btnTest)
        self._vloMainLayout.addWidget(self._sldTest)
        self._vloMainLayout.addWidget(self._pbTest)
        self._vloMainLayout.addWidget(self._pbTest2)

        self.setLayout(self._vloMainLayout)


class ThemeCreator(QDialog):
    """ A dialog for creating a custom widget theme / editing existing ones. """

    def __init__(self) -> None:
        """ Initializer for the class. """

        super().__init__(parent=None)

        self.setWindowTitle("Theme creator")

        self._setup_ui()
        self._setup_connections()

    def _setup_ui(self) -> None:
        """ Sets up the user interface: GUI objects and layouts. """

        # GUI objects
        self._chkUseExistingTheme = QCheckBox(text="Use existing theme",
                                              parent=None)
        self._chkUseExistingTheme.setChecked(True)
        themes = [fn.capitalize().split('.')[0]
                  for fn in os.listdir(_THEME_DIR)]

        self._cmbAvailableThemes = QComboBox()
        self._cmbAvailableThemes.addItems(themes)

        self._fields = "Window WindowText Base AlternateBase ToolTipBase "\
                       "ToolTipText Text Button ButtonText BrightText "\
                       "Highlight HighlightedText".split()
        self._lbllistFields = [QLabel(f) for f in self._fields]
        self._cslist = [_ColourSetter()
                        for _ in range(len(self._fields))]
        self._btnUpdate = QPushButton("Update preview")
        self._btnExport = QPushButton(
            "Export theme (name should be in lowercase)")
        self._btnDelete = QPushButton("Delete theme")
        self._tpPreview = _ThemePreview()

        # Layouts
        self._hloExistingThemes = QHBoxLayout()
        self._hloExistingThemes.addWidget(self._chkUseExistingTheme)
        self._hloExistingThemes.addWidget(self._cmbAvailableThemes)

        self._vloThemeControls = QVBoxLayout()
        self._vloThemeControls.addLayout(self._hloExistingThemes)
        self._hlolistFields = [QHBoxLayout()for _ in range(
            len(self._fields))]
        for hlo, lbl, cs in zip(self._hlolistFields,
                                self._lbllistFields,
                                self._cslist):
            hlo.addWidget(lbl)
            hlo.addStretch(0)
            hlo.addWidget(cs)
            self._vloThemeControls.addLayout(hlo)

        self._vloThemeControls.addWidget(self._btnUpdate)
        self._vloThemeControls.addWidget(self._btnExport)
        self._vloThemeControls.addWidget(self._btnDelete)
        self._vloThemeControls.addStretch(0)

        self._vloPreview = QVBoxLayout()
        self._vloPreview.addWidget(self._tpPreview)
        self._vloPreview.addStretch(0)

        self._hloMainLayout = QHBoxLayout()
        self._hloMainLayout.addLayout(self._vloThemeControls)
        self._hloMainLayout.addLayout(self._vloPreview)

        self.setLayout(self._hloMainLayout)

        # Further initialization
        self._slot_use_existing_theme()
        self._slot_update_by_combobox(0)

    def _setup_connections(self) -> None:
        """ Sets up the connections of the GUI objects. """

        qt_connect(self._chkUseExistingTheme.stateChanged,
                   self._slot_use_existing_theme)
        qt_connect(self._cmbAvailableThemes.currentIndexChanged,
                   self._slot_update_by_combobox)
        qt_connect(self._btnUpdate.clicked, self._slot_update_by_custom_colours)
        qt_connect(self._btnExport.clicked, self._slot_export_theme)
        qt_connect(self._btnDelete.clicked, self._slot_delete_theme)

    def _slot_use_existing_theme(self) -> None:
        """
        Updates the controls' enabled state based on the state of the checkbox.
        """

        use_existing_theme = self._chkUseExistingTheme.isChecked()
        self._cmbAvailableThemes.setEnabled(use_existing_theme)
        for cs in self._cslist:
            cs.set_enabled(not use_existing_theme)

        self._btnUpdate.setEnabled(not use_existing_theme)
        self._btnExport.setEnabled(not use_existing_theme)
        self._btnDelete.setEnabled(use_existing_theme)

        if use_existing_theme:  # To reset to the theme set in the combobox
            theme_idx = self._cmbAvailableThemes.currentIndex()
            self._slot_update_by_combobox(theme_idx)

    def _slot_update_by_combobox(self, index: int) -> None:
        """ Updates the preview based on the selection made in the combobox.

        :param index: The index of the item selected in the combobox.
        """

        theme_name = self._cmbAvailableThemes.itemText(index).lower()
        theme = getattr(WidgetTheme, theme_name)
        for f in fields(theme):
            try:
                field_idx = self._fields.index(f.name)
            except ValueError:
                pass  # Skipping src_path
            else:
                self._cslist[field_idx].colour = getattr(theme, f.name)

        set_widget_theme(self, theme)

    def _slot_update_by_custom_colours(self) -> None:
        """ Updates the preview based on the set custom colours. """

        theme = ThemeParameters()
        for attr, cs in zip(self._fields, self._cslist):
            setattr(theme, attr, cs.colour)

        set_widget_theme(self, theme)

    def _slot_export_theme(self) -> None:
        """
        Exports the currently set custom theme and updates the dialog
        accordingly.
        """

        theme = ThemeParameters()
        colour_attrs = [f.name for f in fields(theme) if f.name != 'src_file']
        for attr, cs in zip(colour_attrs, self._cslist):
            setattr(theme, attr, cs.colour)

        success, path = custom_dialog(self, CFDType.destination_themes)
        if success:
            new_theme = path.split('/')[-1].split('.')[0].capitalize()
            theme.write_json(path)
            WidgetTheme.load_dict()
            with SignalBlocker(self._cmbAvailableThemes) as obj:
                self._cmbAvailableThemes.clear()
                themes = [fn.capitalize().split('.')[0]
                          for fn in os.listdir(_THEME_DIR)]
                obj.addItems(themes)
                obj.setCurrentIndex(themes.index(new_theme))

            self._chkUseExistingTheme.setChecked(True)

    def _slot_delete_theme(self) -> None:
        """ Deletes the currently viewed theme's JSON file and
        updates the dialog accordingly. """

        theme = self._cmbAvailableThemes.currentText().lower()
        os.remove(os.path.join(_THEME_DIR, f'{theme}.json'))
        WidgetTheme.load_dict()
        with SignalBlocker(self._cmbAvailableThemes) as obj:
            self._cmbAvailableThemes.clear()
            themes = [fn.capitalize().split('.')[0]
                      for fn in os.listdir(_THEME_DIR)]
            obj.addItems(themes)
            obj.setCurrentIndex(0)


class _TestApplication(QMainWindow):
    """ The entry point for testing. """

    def __init__(self) -> None:
        """ Initializer for the class. """

        super().__init__(parent=None)

        self.setWindowTitle("Test application")

        # GUI and layouts
        self._setup_ui()
        self._setup_connections()

    def _setup_ui(self) -> None:
        """ Sets up the user interface: GUI objects and layouts. """

        # GUI objects
        self._btnThemeCreator = QPushButton("Open a theme creator dialog")

        # Layouts
        self._vloMainLayout = QVBoxLayout()
        self._vloMainLayout.addWidget(self._btnThemeCreator)

        self._wdgCentralWidget = QWidget()
        self._wdgCentralWidget.setLayout(self._vloMainLayout)
        self.setCentralWidget(self._wdgCentralWidget)

    def _setup_connections(self) -> None:
        """ Sets up the connections of the GUI objects. """

        qt_connect(self._btnThemeCreator.clicked, self._slot_tc_test)

    @classmethod
    def _slot_tc_test(cls) -> None:
        """ Tests the theme creator dialog. """

        tc = ThemeCreator()
        tc.exec()


def _init_module() -> None:
    """ Initializes the module. """

    if not os.path.exists(os.path.join(_PACKAGE_DIR, 'theme_creator.pyi')):
        reprs = []
        class_reprs = []
        classes = {_ColourSetter: None,
                   _ThemePreview: None,
                   ThemeCreator: None,
                   _TestApplication: None}
        for cls, sigs in classes.items():
            add_getattr = not cls.__name__.startswith('_')
            class_reprs.append(stub_repr(cls, signals=sigs,
                                         add_getattr=add_getattr))

        reprs.append('\n\n'.join(class_reprs))

        repr_ = "from typing import Any\n" \
                "from PySide6.QtGui import QColor\n" \
                "from PySide6.QtWidgets import QDialog, QMainWindow, " \
                "QWidget\n\n\n" \
                f"{''.join(reprs)}"

        with open(os.path.join(_PACKAGE_DIR, 'theme_creator.pyi'), 'w') as f:
            f.write(repr_)


_init_module()


if __name__ == '__main__':
    app = QApplication(sys.argv)  # type: ignore
    app.setStyle('Fusion')
    mainWindow = _TestApplication()
    mainWindow.show()
    app.exec()



""" A cancellable, optionally nested progress dialog, reporting on a compatible
process ran by a QObject-subclass on a separate thread. """

__author__ = "Mihaly Konda"
__version__ = '1.0.2'

# Built-in modules
import sys

# Qt6 modules
from PySide6.QtCore import *
from PySide6.QtWidgets import *

# Custom modules
from utils_qt_mk.config import icon_file_path, use_theme
_ICON_FILE_PATH = icon_file_path()
_USE_THEME = use_theme()

from utils_qt_mk.general import qt_connect

if _USE_THEME:
    from utils_qt_mk.theme import set_widget_theme, ThemeParameters, WidgetTheme


def icon_file_path() -> str:
    """ Returns the path for the icon file to be used in the dialogs. """

    return _ICON_FILE_PATH


def set_icon_file_path(new_path: str = '') -> None:
    """ Sets the path for the icon file to be used in the dialogs.

    :param new_path: The new path to set for the windows. The default is an
        empty string, leading to the default icon.
    """

    global _ICON_FILE_PATH
    _ICON_FILE_PATH = new_path


class _Threaded(QObject):
    """
    An example, progress dialog compatible class.
    Worker objects of simple (not nested) PDs do not need sub-signals.

    :cvar sig_new_process_unit: A signal carrying a string identifier of the
        new main process unit.
    :cvar sig_new_subprocess_unit: A signal carrying a string identifier of
        the new subprocess unit.
    :cvar sig_main_progress: A signal carrying the progress (integer percentage)
        of the main process.
    :cvar sig_sub_progress: A signal carrying the progress (integer percentage)
        of the subprocess.
    :cvar sig_finished: The signal emitted when the process finishes.
    :cvar sig_start: The signal that starts the process execution.
    :cvar sig_cancel: A signal for external cancellation of the process
        execution.
    """

    sig_new_process_unit = Signal(str)
    sig_new_subprocess_unit = Signal(str)
    sig_main_progress = Signal(int)
    sig_sub_progress = Signal(int)
    sig_finished = Signal()  # This should carry the returned value of the proc.
    sig_start = Signal()
    sig_cancel = Signal()

    def __init__(self, nested=False) -> None:
        """ Initializer for the class.

        :param nested: A flag marking whether the progress dialog should be
            nested. The default is False.
        """

        super().__init__(parent=None)

        self._nested = nested
        self._canceled = False

        self.sig_start.connect(self._process)
        self.sig_cancel.connect(self._cancel_process)

    @property
    def nested(self) -> bool:
        """ Returns whether the associated progress dialog is nested. """

        return self._nested

    @Slot()
    def _process(self) -> None:
        """
        The slot connected to the start signal.
        Emits `sig_finished` when the process finishes.
        """

        if self._nested:
            for i in range(4):
                self.sig_new_process_unit.emit(f'Outer Iteration {i+1}')
                for j in range(6):
                    self.sig_new_subprocess_unit.emit(f'Inner Iteration {j+1}')
                    self.sig_main_progress.emit((i + 1) * 25)
                    self.sig_sub_progress.emit((j + 1) * 100/6)
                    QThread.msleep(1000)  # type: ignore
                    QCoreApplication.processEvents()  # type: ignore
                    if self._canceled:  # ^^^ To catch cancellation
                        break

                if self._canceled:
                    break
        else:
            for i in range(4):
                self.sig_new_process_unit.emit(f'Iteration {i+1}')
                self.sig_main_progress.emit((i + 1) * 25)
                QThread.msleep(1000)  # type: ignore
                QCoreApplication.processEvents()  # type: ignore
                if self._canceled:  # ^^^ To catch cancellation
                    break

        self.sig_finished.emit()

    @Slot()
    def _cancel_process(self) -> None:
        """
        The slot connected to the cancel signal.
        Needs the button press event to be processed beforehand.
        """

        self._canceled = True


class _ProgressMixin:
    """
    A window reporting on the progress of a process running on a separate
    thread.
    """

    def __init__(self, worker, title, widget_theme=None) -> None:
        """ Initializer for the class.

        :param worker: A worker subclassing QObject, handling a process.
        :param title: The title to set for the window.
        :param widget_theme: A widget theme from the 'theme' module. The default
            is None, for the Qt-default theme.
        """

        super().__init__()

        close_removed = (self.windowFlags().value -  # type: ignore
                         Qt.WindowType.WindowCloseButtonHint.value)
        self.setWindowFlags(Qt.WindowType(close_removed))  # type: ignore
        self.setWindowTitle(title)  # type: ignore
        if _ICON_FILE_PATH:
            self.setWindowIcon(QIcon(_ICON_FILE_PATH))  # type: ignore

        self._worker = worker
        self._widget_theme = widget_theme

        self._setup_ui()
        self._setup_connections()

        self._create_worker_thread()

    def _setup_ui(self) -> None:
        """ Sets up the user interface: GUI objects and layouts. """

        # GUI objects
        self._lblMain = QLabel(parent=None)
        self._pbMain = QProgressBar()
        self._pbMain.setFixedWidth(500)
        self._lblSub = QLabel(parent=None)
        self._pbSub = QProgressBar()
        self._pbSub.setFixedWidth(500)
        self._btnCancel = QPushButton('Cancel')

        # Layouts
        self._vloMainLayout = QVBoxLayout()
        self._vloMainLayout.addWidget(self._lblMain)
        self._vloMainLayout.addWidget(self._pbMain)
        self._vloMainLayout.addWidget(self._lblSub)
        self._vloMainLayout.addWidget(self._pbSub)
        self._vloMainLayout.addWidget(self._btnCancel)
        self.setLayout(self._vloMainLayout)  # type: ignore

        # Further initializations
        if not self._worker.nested:
            self._lblSub.hide()
            self._pbSub.hide()

        if _USE_THEME:
            set_widget_theme(self)  # type: ignore

    def _setup_connections(self) -> None:
        """ Sets up the connections of the GUI objects. """

        qt_connect(self._worker.sig_new_process_unit, self._lblMain.setText)
        qt_connect(self._worker.sig_main_progress, self._pbMain.setValue)

        if self._worker.nested:  # The worker shouldn't have these if not nested
            qt_connect(self._worker.sig_new_subprocess_unit,
                       self._lblSub.setText)
            qt_connect(self._worker.sig_sub_progress, self._pbSub.setValue)

        qt_connect(self._worker.sig_finished, self._quit_thread)
        qt_connect(self._btnCancel.clicked, self._cancel_process)

    @property
    def theme(self) -> ThemeParameters:
        """ Returns the parameters of the theme set for this object. """

        return self._widget_theme

    @theme.setter
    def theme(self, new_theme: ThemeParameters) -> None:
        """ Sets a new set of parameters defining a theme to this object. """

        self._widget_theme = new_theme

    def _create_worker_thread(self) -> None:
        """ Creates a new thread and moves the worker there. """

        self._worker_thread = QThread()  # type: ignore
        self._worker_thread.start()
        self._worker.moveToThread(self._worker_thread)
        self._worker.sig_start.emit()

    def _quit_thread(self) -> None:
        """ Closes the thread after the worker has finished. """

        self._worker_thread.quit()
        self._worker_thread.wait()
        self.close()  # type: ignore

    def _cancel_process(self) -> None:
        """ Asks for confirmation to cancel the process. """

        reply = QMessageBox.question(self, 'Cancel Process',  # type: ignore
                                     "Are you sure you want to cancel "
                                     "the process?",
                                     QMessageBox.StandardButton.Yes |
                                     QMessageBox.StandardButton.No,
                                     QMessageBox.StandardButton.No)
        if reply == QMessageBox.StandardButton.Yes:
            self._worker.sig_cancel.emit()


class ProgressDialog(_ProgressMixin, QDialog):
    """
    A dialog reporting on the progress of a process running on a separate
    thread.
    """

    def __init__(self, worker: QObject, title="Progress report",
                 widget_theme=None) -> None:
        """ Initializer for the class.

        :param worker: A worker subclassing QObject, handling a process.
        :param title: The title to set for the dialog. The default is
            "Progress report".
        :param widget_theme: A widget theme from the 'theme' module. The default
            is None, for the Qt-default theme.
        """

        super().__init__(worker, title, widget_theme)


class ProgressDW(_ProgressMixin, QDockWidget):
    """
    A dock widget reporting on the progress of a process running on a separate
    thread.
    """

    def __init__(self, worker: QObject, title="Progress report",
                 widget_theme=None) -> None:
        """ Initializer for the class.

        :param worker: A worker subclassing QObject, handling a process.
        :param title: The title to set for the dialog. The default is
            "Progress report".
        :param widget_theme: A widget theme from the 'theme' module. The default
            is None, for the Qt-default theme.
        """

        super().__init__(worker, title, widget_theme)

        self._wdgContent = QWidget()  # type: ignore
        self._wdgContent.setLayout(self._vloMainLayout)
        self.setWidget(self._wdgContent)
        self.setAllowedAreas(Qt.DockWidgetArea.AllDockWidgetAreas)
        self.setFloating(True)


class _TestApplication(QMainWindow):
    """ The entry point for testing. """

    def __init__(self) -> None:
        """ Initializer for the class. """

        super().__init__(parent=None)

        self.setWindowTitle("Test application")

        # GUI and layouts
        self._setup_ui()
        self._setup_connections()

    def _setup_ui(self) -> None:
        """ Sets up the user interface: GUI objects and layouts. """

        # GUI objects
        self._btnToggleTheme = QPushButton("Toggle theme (enabled)")
        self._btnSimplePD = QPushButton("Open a simple progress dialog")
        self._btnSimplePD.setObjectName('simple_d')
        self._btnSimplePDW = QPushButton("Open a simple progress DW")
        self._btnSimplePDW.setObjectName('simple_dw')
        self._btnNestedPD = QPushButton("Open a nested process dialog")
        self._btnNestedPD.setObjectName('nested_d')
        self._btnNestedPDW = QPushButton("Open a nested process DW")
        self._btnNestedPDW.setObjectName('nested_dw')

        # Layouts
        self._vloMainLayout = QVBoxLayout()
        self._vloMainLayout.addWidget(self._btnToggleTheme)
        self._vloMainLayout.addWidget(self._btnSimplePD)
        self._vloMainLayout.addWidget(self._btnSimplePDW)
        self._vloMainLayout.addWidget(self._btnNestedPD)
        self._vloMainLayout.addWidget(self._btnNestedPDW)

        self._wdgCentralWidget = QWidget()
        self._wdgCentralWidget.setLayout(self._vloMainLayout)
        self.setCentralWidget(self._wdgCentralWidget)

    def _setup_connections(self) -> None:
        """ Sets up the connections of the GUI objects. """

        qt_connect(self._btnToggleTheme.clicked, self._slot_toggle_theme)
        qt_connect(self._btnSimplePD.clicked, self._slot_test)
        qt_connect(self._btnNestedPD.clicked, self._slot_test)
        qt_connect(self._btnSimplePDW.clicked, self._slot_test)
        qt_connect(self._btnNestedPDW.clicked, self._slot_test)

    def _slot_toggle_theme(self) -> None:
        """ Unlocks the theme module to test the theming of the PD. """

        global _USE_THEME
        _USE_THEME = not _USE_THEME
        self._btnToggleTheme.setText(f"Toggle theme ("
                                     f"{'enabled' if _USE_THEME
                                        else 'disabled'})")

    def _slot_test(self) -> None:
        """ Tests the progress dialogs/dock widgets. """

        def catch_signal() -> None:
            """
            Catches the finished signal of the worker object.
            Also emitted when the process is cancelled.
            """

            print("Worker object's process is finished!")

        wo = _Threaded('nested' in self.sender().objectName())
        wo.sig_finished.connect(catch_signal)
        theme = None if not _USE_THEME else WidgetTheme.yellow
        if 'dw' in self.sender().objectName():
            self._test = ProgressDW(wo, "Custom title", theme)
        else:
            self._test = ProgressDialog(wo, "Custom title", theme)

        starter = 'show' if 'dw' in self.sender().objectName() else 'exec'
        self._test.setWindowModality(Qt.WindowModality.ApplicationModal)
        getattr(self._test, starter)()


if __name__ == '__main__':
    app = QApplication(sys.argv)  # type: ignore
    app.setStyle('Fusion')
    mainWindow = _TestApplication()
    mainWindow.show()
    app.exec()



from dataclasses import dataclass
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDialog, QMainWindow, QMessageBox, QWidget
from utils_qt_mk._general import Singleton


MessageBoxType: _MessageBoxType = None
_MBCategories: _MessageBoxCategories = None
_StandardButtons: dict[int, QMessageBox.StandardButton] = None
_WindowTypes: dict[int, Qt.WindowType] = None

def get_messagebox_types(fetch_data: bool = False) -> list[str | _MessageBoxData]: ...
def merge_json(path: str) -> None: ...
def message(parent: QWidget, mbd: _MessageBoxData, custom_text: str = None) -> QMessageBox.StandardButton: ...


@dataclass
class _MessageBoxData:
	def __init__(self, icon: QMessageBox.Icon = Icon.NoIcon, title: str = '', text: str = '', buttons: list[QMessageBox.StandardButton] = None, flags: list[Qt.WindowType] = None) -> None: ...
	@classmethod
	def from_dict(cls, src: dict) -> _MessageBoxData: ...
	def merged_bits(self, attr: str) -> QMessageBox.StandardButton | Qt.WindowType: ...
	@property
	def as_dict(self) -> dict: ...


@dataclass
class _MessageBoxCategories(metaclass=Singleton):
	def __init__(self) -> None: ...


class _MessageBoxType(metaclass=Singleton):
	close_event_question: _MessageBoxData = None
	def __init__(self) -> None: ...
	def converted_keys(self) -> list[str]: ...
	def export_types(self) -> None: ...
	def import_types(self) -> None: ...
	def is_empty(self) -> bool: ...


class _OrderedSelectionList(QWidget):
	def __init__(self, list_name: str, items: list, add: str, remove: str) -> None: ...
	def _setup_connections(self) -> None: ...
	def _setup_ui(self) -> None: ...
	def _slot_add_item(self) -> None: ...
	def _slot_remove_item(self) -> None: ...
	def setEnabled(self, new_state: bool) -> None: ...
	def set_selection(self, new_selection: list[str]) -> None: ...
	@property
	def selection_idx(self) -> list[int]: ...
	@property
	def selection_str(self) -> list[str]: ...


class _MessageBoxTypeCreator(QDialog):
	def __init__(self) -> None: ...
	def _get_as_messageboxdata(self) -> _MessageBoxData: ...
	def _setup_connections(self) -> None: ...
	def _setup_ui(self) -> None: ...
	def _slot_delete_settings(self) -> None: ...
	def _slot_export_settings(self) -> None: ...
	def _slot_set_control_states(self) -> None: ...
	def _slot_test_settings(self) -> None: ...
	def _slot_update_by_combobox(self) -> None: ...


class _TestApplication(QMainWindow):
	def __init__(self) -> None: ...
	def _setup_connections(self) -> None: ...
	def _setup_ui(self) -> None: ...
	@classmethod
	def _slot_mbtc_test(cls) -> None: ...



""" Customizable messageboxes with themes. """

from __future__ import annotations

__author__ = "Mihaly Konda"
__version__ = '1.0.4'


# Built-in modules
from dataclasses import dataclass, field, fields
import json
import os
import sys
from typing import Any, cast, Iterable

# Qt6 modules
from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

# Custom classes/modules
from utils_qt_mk.config import _PACKAGE_DIR, icon_file_path, theme_dir
_ICON_FILE_PATH = icon_file_path()
_THEME_DIR = theme_dir()

from utils_qt_mk.general import SignalBlocker, Singleton, stub_repr, qt_connect
from utils_qt_mk.theme import set_widget_theme, WidgetTheme


MessageBoxType: _MessageBoxType | None = None
_MBCategories: _MessageBoxCategories | None = None
_StandardButtons: dict[int, QMessageBox.StandardButton] = \
    {idx: btn for idx, btn
     in enumerate(cast(Iterable[QMessageBox.StandardButton],
                       QMessageBox.StandardButton))}
_WindowTypes: dict[int, Qt.WindowType] = \
    {idx: typ for idx, typ
     in enumerate(cast(Iterable[Qt.WindowType], Qt.WindowType))}


def icon_file_path() -> str:
    """ Returns the path for the icon file to be used in the dialogs. """

    return _ICON_FILE_PATH


def set_icon_file_path(new_path: str = '') -> None:
    """ Sets the path for the icon file to be used in the dialogs.

    :param new_path: The new path to set for the windows. The default is an
        empty string, leading to the default icon.
    """

    global _ICON_FILE_PATH
    _ICON_FILE_PATH = new_path


def get_messagebox_types(fetch_data: bool = False) \
        -> list[str | _MessageBoxData]:
    """ Returns the available messagebox types.

    :param fetch_data: A flag requesting the _MessageBoxData objects themselves.
    The default is False.
    """

    if fetch_data:
        return [pd for pd in MessageBoxType._types.values()]
    else:
        return [key.lower() for key in MessageBoxType._types.keys()]


def merge_json(path: str) -> None:
    """
    Takes an external JSON messagebox type file and merges its contents to
    the package's own file. This way if you create messageboxes you can reuse
    them in another project.

    :param path: Path to the external JSON messagebox type file.
    """

    with open(os.path.join(_PACKAGE_DIR, 'messagebox_types.json'),
              'r') as f:
        package_data = json.load(f)

    package_data = {mbd['type_id']: _MessageBoxData.from_dict(mbd)
                    for mbd in package_data}

    with open(path, 'r') as f:
        external_data = json.load(f)

    external_data = {mbd['type_id']: _MessageBoxData.from_dict(mbd)
                     for mbd in external_data}

    for type_id, mbd_e in external_data.items():
        if all(mbd_e != mbd_p for mbd_p in package_data):
            package_data[type_id] = mbd_e

    package_data = [{'type_id': t_id, **pd.as_dict}
                    for t_id, pd in package_data.items()]

    with open(os.path.join(_PACKAGE_DIR, 'messagebox_types.json'),
              'w') as f:
        json.dump(package_data, f, indent=4)

    try:
        os.remove(os.path.join(_PACKAGE_DIR, 'message.pyi'))
    except FileNotFoundError:
        pass
    else:
        _init_module()  # Reinitialize the module so that types are reloaded


@dataclass
class _MessageBoxData:
    """ Settings defining the appearance of a QMessageBox.

    :param icon: The icon to set for the message box.
    :param title: The title to set for the message box window.
    :param text: The text content to set for the message box.
    :param buttons: A list of standard buttons to set for the message box.
    :param flags: A list of flags to set for the message box defining its
        behaviour.
    """

    icon: QMessageBox.Icon = QMessageBox.Icon.NoIcon
    title: str = ''
    text: str = ''
    buttons: list[QMessageBox.StandardButton] = None
    flags: list[Qt.WindowType] = None

    def __post_init__(self) -> None:
        """ Adds the correct default values where they are mutable. """

        if self.buttons is None:
            self.buttons = [QMessageBox.StandardButton.NoButton]

        if self.flags is None:
            self.flags = [Qt.WindowType.Dialog,
                          Qt.WindowType.MSWindowsFixedSizeDialogHint]

    def __eq__(self, other) -> bool:
        """ Custom comparison rule, comparing each field. """

        if not isinstance(other, _MessageBoxData):
            return False

        return all(getattr(self, f.name) == getattr(other, f.name)
                   for f in fields(self))

    def merged_bits(self, attr: str) \
            -> QMessageBox.StandardButton | Qt.WindowType:
        """ Merges the bits of either `buttons` or `flags` and returns.

        :param attr: The requested attribute (`buttons` or `flags`) as a string.

        :returns: An Enum subclass (concrete type based on `attr`) with the
            merged value.
        """

        bit_pattern = getattr(self, attr)
        ret_types = {'buttons': QMessageBox.StandardButton,
                     'flags': Qt.WindowType}

        merged = 0
        for bp in bit_pattern:
            merged |= bp

        return ret_types[attr](merged)

    @property
    def as_dict(self) -> dict:
        """ Returns the data content as a dictionary. """

        return {'icon': self.icon.value,
                'title': self.title,
                'text': self.text,
                'buttons': [btn.value for btn in self.buttons],
                'flags': [flag.value for flag in self.flags]}

    @classmethod
    def from_dict(cls, src: dict) -> _MessageBoxData:
        """ Returns an instance built from a dictionary.

        :param src: A dictionary containing data to build an instance,
            extracted from the handled JSON file.
        """

        buttons = [QMessageBox.StandardButton(id_) for id_ in src['buttons']]
        flags = [Qt.WindowType(id_) for id_ in src['flags']]

        return cls(QMessageBox.Icon(src['icon']), src['title'], src['text'],
                   buttons, flags)


@dataclass
class _MessageBoxCategories(metaclass=Singleton):  # Not Enum because...
    """
    A constant dataclass for holding parameters of the four basic
    categories of message boxes and an additional custom type.
    """

    critical: _MessageBoxData = field(init=False)  # ...the values are mutable
    information: _MessageBoxData = field(init=False)
    question: _MessageBoxData = field(init=False)
    warning: _MessageBoxData = field(init=False)
    custom: _MessageBoxData = field(init=False)

    def __post_init__(self) -> None:
        """ Creates mutable values after initialization. """

        self.critical = _MessageBoxData(QMessageBox.Icon.Critical,
                                        buttons=[QMessageBox.StandardButton.Ok])
        self.information = _MessageBoxData(QMessageBox.Icon.Information,
                                           buttons=
                                           [QMessageBox.StandardButton.Ok])
        self.question = _MessageBoxData(QMessageBox.Icon.Question,
                                        buttons=[QMessageBox.StandardButton.Yes,
                                                 QMessageBox.StandardButton.No])
        self.warning = _MessageBoxData(QMessageBox.Icon.Warning,
                                       buttons=[QMessageBox.StandardButton.Ok])
        self.custom = _MessageBoxData()


class _MessageBoxType(metaclass=Singleton):
    """ A collection of predefined types of messagebox. """

    def __init__(self) -> None:
        """ Initializer for the class. """

        self._types: dict[str, _MessageBoxData] | None = None
        self.import_types()

    def __getattr__(self, name: str) -> Any:
        """ Handles an attribute access request.

        :param name: The name of the requested attribute.

        :returns: A stored `_MessageBoxData` object or an attribute of the
            internal dictionary.
        """

        try:
            return getattr(self._types, name)  # dict attributes
        except AttributeError:
            return self._types[name]  # _MessageBoxData object

    def __setattr__(self, key: str, value: _MessageBoxData | None) -> None:
        """ Handles an attribute setting request.

        :param key: The name of the attribute whose value is to be set.
        :param value: The value to set for the attribute.
        """

        if key.startswith('_'):  # Avoiding infinite recursion with _types
            dict.__setattr__(self, key, value)
        else:
            self._types[key] = value

    def __setitem__(self, key: str, value: _MessageBoxData) -> None:
        """
        Sets a new set of message box data for the internal dictionary by
        accessing with '[]'.

        :param key: The type ID to set the data to.
        :param value: The message box data to set.
        """

        try:
            self._types[key] = value
        except TypeError:
            self._types = {key: value}

    def __delitem__(self, key: str) -> None:
        """ Deletes a set of message box data from the internal dictionary. """

        del self._types[key]
        if not self._types:
            self._types = None

    def import_types(self) -> None:
        """ Imports types from the handled JSON file. """

        try:
            with (open(os.path.join(_PACKAGE_DIR, 'messagebox_types.json'), 'r')
                  as f):
                data: list[dict] = json.load(f)

            self._types = {}
            for entry in data:
                type_id = entry.pop('type_id')
                self._types[type_id] = _MessageBoxData.from_dict(entry)
        except FileNotFoundError:
            pass

    def export_types(self) -> None:
        """ Exports types to the handled JSON file. """

        data = []
        for type_id, type_data in self._types.items():
            data.append({'type_id': type_id, **type_data.as_dict})

        with (open(os.path.join(_PACKAGE_DIR, 'messagebox_types.json'), 'w')
              as f):
            json.dump(data, f, indent=4)

    def is_empty(self) -> bool:
        """ Returns True if there are no defined types, False if there are. """

        return self._types is None or not self._types

    def converted_keys(self) -> list[str]:
        """
        Returns the keys converted to a list of space-separated and capitalized
        strings.
        """

        keys = self._types.keys()
        return [k.capitalize().replace('_', ' ') for k in keys]


class _OrderedSelectionList(QWidget):
    """ A widget where an ordered selection can be made from a combobox. """

    def __init__(self, list_name: str, items: list, add: str, remove: str) \
            -> None:
        """ Initializer for the class.

        :param list_name: String identifier of the list set to a label.
        :param items: A list of items to set for the combobox.
        :param add: Text to set for the add button.
        :param remove: Text to set for the remove button.
        """

        super().__init__(parent=None)

        self._list_name = list_name
        self._items = {item: idx for idx, item in enumerate(items)}
        self._add = add
        self._remove = remove

        self._setup_ui()
        self._setup_connections()

    def _setup_ui(self) -> None:
        """ Sets up the user interface: GUI objects and layouts. """

        # GUI objects
        self._lwSelection = QListWidget()
        self._lwSelection.setDragDropMode(
            QAbstractItemView.DragDropMode.InternalMove)

        self._lblList = QLabel(text=self._list_name, parent=None)
        self._cmbItems = QComboBox()
        self._cmbItems.addItems(self._items.keys())  # type: ignore
        self._btnAdd = QPushButton(self._add)
        self._btnRemove = QPushButton(self._remove)

        # Layouts
        self._vloControls = QVBoxLayout()
        self._vloControls.addWidget(self._lblList)
        self._vloControls.addWidget(self._cmbItems)
        self._vloControls.addWidget(self._btnAdd)
        self._vloControls.addWidget(self._btnRemove)
        self._vloControls.addStretch(0)

        self._hloMainLayout = QHBoxLayout()
        self._hloMainLayout.addWidget(self._lwSelection)
        self._hloMainLayout.addLayout(self._vloControls)

        self.setLayout(self._hloMainLayout)

    def _setup_connections(self) -> None:
        """ Sets up the connections of the GUI objects. """

        qt_connect(self._btnAdd.clicked, self._slot_add_item)
        qt_connect(self._btnRemove.clicked, self._slot_remove_item)

    def _slot_add_item(self) -> None:
        """ Adds the current item of the combobox to the selection list. """

        new_item = self._cmbItems.currentText()
        if new_item not in self.selection_str:
            self._lwSelection.addItem(new_item)

    def _slot_remove_item(self) -> None:
        """ Removes the currently selected item from the selection list. """

        if (idx := self._lwSelection.currentIndex().row()) >= 0:
            self._lwSelection.takeItem(idx)

    def set_selection(self, new_selection: list[str]) -> None:
        """ Resets the selection list by the provided items.

        :param new_selection: The new items to set to the selection list.
        """

        self._lwSelection.clear()
        for item in new_selection:
            if item in self._items.keys():
                self._lwSelection.addItem(item)

    @property
    def selection_str(self) -> list[str]:
        """ Returns the string content of the selection list. """

        return [self._lwSelection.item(i).text()
                for i in range(self._lwSelection.count())]

    @property
    def selection_idx(self) -> list[int]:
        """
        Returns the selection list encoded by the order of the source item list.
        """

        return [self._items[s] for s in self.selection_str]

    def setEnabled(self, new_state: bool) -> None:
        """ Sets the enabled state of the child widgets.

        :param new_state: The new enabled state to set.
        """

        self._lwSelection.setEnabled(new_state)
        self._cmbItems.setEnabled(new_state)
        self._btnAdd.setEnabled(new_state)
        self._btnRemove.setEnabled(new_state)


class _MessageBoxTypeCreator(QDialog):
    """
    A dialog for defining custom messagebox types / editing existing ones.
    """

    def __init__(self) -> None:
        """ Initializer for the class. """

        super().__init__(parent=None)

        self.setWindowTitle("MBT Creator")

        self._categories = ("critical information question "
                            "warning custom".split())

        self._setup_ui()
        self._setup_connections()

    def _setup_ui(self) -> None:
        """ Sets up the user interface: GUI objects and layouts. """

        # GUI objects
        self._chkUseExistingType = QCheckBox(text="Use existing type",
                                             parent=None)
        self._chkUseExistingType.setEnabled(not MessageBoxType.is_empty())
        self._chkUseExistingType.setObjectName('checkbox')

        self._cmbAvailableTypes = QComboBox()
        self._cmbAvailableTypes.setObjectName('types')
        if not MessageBoxType.is_empty():
            self._cmbAvailableTypes.addItems(MessageBoxType.converted_keys())

        self._ledTypeID = QLineEdit()
        self._ledTypeID.setPlaceholderText("Type ID")

        self._lblCategory = QLabel(text='Category', parent=None)
        self._cmbCategories = QComboBox()
        self._cmbCategories.setObjectName('categories')
        self._cmbCategories.addItems(self._categories)
        self._cmbCategories.setObjectName('combobox')

        self._lblIcon = QLabel(text='Icon', parent=None)
        self._cmbIcons = QComboBox()
        self._cmbIcons.addItems([icon.name for icon in QMessageBox.Icon])

        self._ledTitle = QLineEdit()
        self._ledTitle.setPlaceholderText("Window title")
        self._tedText = QTextEdit()
        self._tedText.setPlaceholderText('Message')

        buttons = [btn.name for btn in _StandardButtons.values()]
        self._oslButtons = _OrderedSelectionList('Buttons',
                                                 buttons,
                                                 "Add button",
                                                 "Remove button")

        flags = [wt.name for wt in Qt.WindowType]
        self._oslFlags = _OrderedSelectionList('Flags',
                                               flags,
                                               "Add flag",
                                               "Remove flag")

        self._btnTest = QPushButton("Show test message")
        self._btnExport = QPushButton("Export type")
        self._btnDelete = QPushButton("Delete type")

        # Layouts
        self._hloExistingTypes = QHBoxLayout()
        self._hloExistingTypes.addWidget(self._chkUseExistingType)
        self._hloExistingTypes.addWidget(self._cmbAvailableTypes)

        self._hloCategory = QHBoxLayout()
        self._hloCategory.addWidget(self._lblCategory)
        self._hloCategory.addWidget(self._cmbCategories)

        self._hloIcon = QHBoxLayout()
        self._hloIcon.addWidget(self._lblIcon)
        self._hloIcon.addWidget(self._cmbIcons)

        self._vloMainLayout = QVBoxLayout()
        self._vloMainLayout.addLayout(self._hloExistingTypes)
        self._vloMainLayout.addWidget(self._ledTypeID)
        self._vloMainLayout.addLayout(self._hloCategory)
        self._vloMainLayout.addLayout(self._hloIcon)
        self._vloMainLayout.addWidget(self._ledTitle)
        self._vloMainLayout.addWidget(self._tedText)
        self._vloMainLayout.addWidget(self._oslButtons)
        self._vloMainLayout.addWidget(self._oslFlags)
        self._vloMainLayout.addWidget(self._btnTest)
        self._vloMainLayout.addWidget(self._btnExport)
        self._vloMainLayout.addWidget(self._btnDelete)
        self._vloMainLayout.addStretch(0)

        self.setLayout(self._vloMainLayout)

        # Further initialization
        self._slot_set_control_states()

    def _setup_connections(self) -> None:
        """ Sets up the connections of the GUI objects. """

        qt_connect(self._chkUseExistingType.stateChanged,
                   self._slot_set_control_states)
        qt_connect(self._cmbAvailableTypes.currentIndexChanged,
                   self._slot_update_by_combobox)
        qt_connect(self._cmbCategories.currentIndexChanged,
                   self._slot_set_control_states)
        qt_connect(self._btnTest.clicked, self._slot_test_settings)
        qt_connect(self._btnExport.clicked, self._slot_export_settings)
        qt_connect(self._btnDelete.clicked, self._slot_delete_settings)

    def _slot_set_control_states(self) -> None:
        """ Updates the controls' enabled state based on the state of
                the checkbox. """

        use_existing_type = self._chkUseExistingType.isChecked()
        standard = self._cmbCategories.currentText() != 'custom'

        self._cmbAvailableTypes.setEnabled(use_existing_type)
        self._ledTypeID.setVisible(not use_existing_type)
        self._cmbCategories.setEnabled(not use_existing_type)

        self._cmbIcons.setEnabled(not use_existing_type and not standard)
        self._ledTitle.setEnabled(not use_existing_type)
        self._tedText.setEnabled(not use_existing_type)
        self._oslButtons.setEnabled(not use_existing_type and not standard)
        self._oslFlags.setEnabled(not use_existing_type and not standard)

        self._btnExport.setEnabled(not use_existing_type)
        self._btnDelete.setEnabled(use_existing_type)

        if not use_existing_type:
            mbd = getattr(_MBCategories, self._cmbCategories.currentText())
        else:
            key = (self._cmbAvailableTypes.currentText()
                   .lower().replace(' ', '_'))
            mbd = getattr(MessageBoxType, key)

        self._cmbIcons.setCurrentIndex(mbd.icon.value)
        self._ledTitle.setText(mbd.title)
        self._tedText.setText(mbd.text)
        self._oslButtons.set_selection([btn.name for btn in mbd.buttons])
        self._oslFlags.set_selection([flag.name for flag in mbd.flags])

    def _slot_update_by_combobox(self) -> None:
        """ Updates the dialog according to the controlling combobox. """

        typ = self._cmbAvailableTypes.currentText().lower().replace(' ', '_')
        mbd: _MessageBoxData = getattr(MessageBoxType, typ)

        self._cmbIcons.setCurrentIndex(mbd.icon.value)
        self._ledTitle.setText(mbd.title)
        self._tedText.setText(mbd.text)
        self._oslButtons.set_selection(
            [btn.name for btn in mbd.buttons])  # type: ignore
        self._oslFlags.set_selection(
            [f.name for f in mbd.flags])  # type: ignore

    def _get_as_messageboxdata(self) -> _MessageBoxData:
        """
        Returns a MessageBoxData object built from the settings made in the
        dialog.
        """

        buttons = [_StandardButtons[idx]
                   for idx in self._oslButtons.selection_idx]
        flags = [_WindowTypes[idx] for idx in self._oslFlags.selection_idx]
        return _MessageBoxData(QMessageBox.Icon(self._cmbIcons.currentIndex()),
                               self._ledTitle.text(),
                               self._tedText.toPlainText(),
                               buttons,
                               flags)

    def _slot_test_settings(self) -> None:
        """ Creates a message box dialog based on the settings. """

        retval = message(self, self._get_as_messageboxdata())
        print(f"The message box returned {retval} "
              f"({QMessageBox.StandardButton(retval).name}).")

    def _slot_export_settings(self) -> None:
        """
        Exports the currently set type and updates the dialog accordingly.
        """

        if self._chkUseExistingType.isChecked():
            type_id = (self._cmbAvailableTypes.currentText().lower()
                       .replace(' ', '_'))
        else:
            type_id = self._ledTypeID.text().lower().replace(' ', '_')

        # Data updated, no need to reimport
        MessageBoxType[type_id] = self._get_as_messageboxdata()
        MessageBoxType.export_types()
        with SignalBlocker(self._cmbAvailableTypes) as obj:
            obj.clear()
            obj.addItems(MessageBoxType.converted_keys())
            obj.setCurrentIndex(obj.count() - 1)

        self._chkUseExistingType.setEnabled(True)

    def _slot_delete_settings(self) -> None:
        """
        Deletes the currently selected type and updates the dialog accordingly.
        """

        type_id = (self._cmbAvailableTypes.currentText().lower()
                   .replace(' ', '_'))
        del MessageBoxType[type_id]
        with SignalBlocker(self._cmbAvailableTypes) as obj:
            obj.clear()
            if MessageBoxType.is_empty():
                os.remove('messagebox_types.json')
            else:
                MessageBoxType.export_types()
                obj.addItems(MessageBoxType.converted_keys())
                obj.setCurrentIndex(obj.count() - 1)

        self._slot_update_by_combobox()  # One update after signal got unblocked

        if MessageBoxType.is_empty():
            with SignalBlocker(self._chkUseExistingType) as obj:
                obj.setEnabled(False)


def message(parent: QWidget, mbd: _MessageBoxData, custom_text: str = None) \
        -> QMessageBox.StandardButton:
    """
    Shows a modal QMessageBox with preset content (or custom text) and a custom
    theme.

    :param parent: The parent widget calling for the message dialog.
    :param mbd: MessageBox data to define the appearance of the created window.
    :param custom_text: Overrides the preset text. The default is None, having
        no effect.

    :returns: The clicked standard button (or its equal if the dialog was just
        closed).
    """

    default = os.listdir(_THEME_DIR)[0].split('/')[-1].split('.')[0]
    theme = getattr(WidgetTheme, default)

    try:
        theme = parent.theme  # type: ignore
    except AttributeError:
        print(f"Cannot access the theme of the parent object of class "
              f"'{parent.__class__.__name__}' or it has no theme. "
              f"Using the default theme ({default}).")

    text = mbd.text if custom_text is None else custom_text
    messagebox = QMessageBox(mbd.icon, mbd.title, text,
                             mbd.merged_bits('buttons'), parent,
                             mbd.merged_bits('flags'))

    if _ICON_FILE_PATH:
        messagebox.setWindowIcon(QIcon(_ICON_FILE_PATH))  # type: ignore

    set_widget_theme(messagebox, theme)
    messagebox.setWindowModality(Qt.WindowModality.ApplicationModal)
    return QMessageBox.StandardButton(messagebox.exec())


class _TestApplication(QMainWindow):
    """ The entry point for testing. """

    def __init__(self) -> None:
        """ Initializer for the class. """

        super().__init__(parent=None)

        self.setWindowTitle("Test application")

        # GUI and layouts
        self._setup_ui()
        self._setup_connections()

    def _setup_ui(self) -> None:
        """ Sets up the user interface: GUI objects and layouts. """

        # GUI objects
        self._btnMBTCreator = QPushButton("Open a MessageBoxType "
                                          "creator dialog")

        # Layouts
        self._vloMainLayout = QVBoxLayout()
        self._vloMainLayout.addWidget(self._btnMBTCreator)

        self._wdgCentralWidget = QWidget()
        self._wdgCentralWidget.setLayout(self._vloMainLayout)
        self.setCentralWidget(self._wdgCentralWidget)

    def _setup_connections(self) -> None:
        """ Sets up the connections of the GUI objects. """

        qt_connect(self._btnMBTCreator.clicked, self._slot_mbtc_test)

    @classmethod
    def _slot_mbtc_test(cls) -> None:
        """ Tests the MessageBoxType creator dialog. """

        mbtc = _MessageBoxTypeCreator()
        mbtc.exec()


def _init_module() -> None:
    """ Initializes the module. """

    if not os.path.exists(os.path.join(_PACKAGE_DIR, 'message.pyi')):
        functions = [get_messagebox_types, merge_json, message]
        reprs = [stub_repr(func) for func in functions]
        reprs.append('\n\n')
        class_reprs = []
        classes = {_MessageBoxData: None,
                   _MessageBoxCategories: None,
                   _MessageBoxType: None,
                   _OrderedSelectionList: None,
                   _MessageBoxTypeCreator: None,
                   _TestApplication: None}
        for cls, sigs in classes.items():
            if cls == _MessageBoxType:
                try:
                    with open(os.path.join(_PACKAGE_DIR,
                                           'messagebox_types.json'), 'r') as f:
                        data: list[dict] = json.load(f)

                    extra_cvs = '\n'.join(
                        [f"\t{entry['type_id']}: _MessageBoxData "
                         "= None" for entry in data])
                except FileNotFoundError:
                    extra_cvs = None
            else:
                extra_cvs = None

            class_reprs.append(
                stub_repr(cls, signals=sigs, extra_cvs=extra_cvs))

        reprs.append('\n\n'.join(class_reprs))

        imports = "from dataclasses import dataclass\n" \
                  "from PySide6.QtCore import Qt\n" \
                  "from PySide6.QtWidgets import QDialog, QMainWindow, "\
                  "QMessageBox, QWidget\n" \
                  "from utils_qt_mk._general import Singleton\n\n\n"

        with open(os.path.join(_PACKAGE_DIR, 'message.pyi'), 'w') as f:
            f.write(imports)
            f.write("MessageBoxType: _MessageBoxType = None\n")
            f.write("_MBCategories: _MessageBoxCategories = None\n")
            f.write("_StandardButtons: dict[int, QMessageBox.StandardButton] "
                    "= None\n")
            f.write("_WindowTypes: dict[int, Qt.WindowType] = None\n\n")
            f.write(''.join(reprs))

    global MessageBoxType
    MessageBoxType = _MessageBoxType()

    global _MBCategories
    _MBCategories = _MessageBoxCategories()


_init_module()


if __name__ == '__main__':
    app = QApplication(sys.argv)  # type: ignore
    app.setStyle('Fusion')
    mainWindow = _TestApplication()
    mainWindow.show()
    app.exec()




""" File dialogs customized by a JSON file. """

from __future__ import annotations

__author__ = "Mihaly Konda"
__version__ = '1.0.7'


# Built-in modules
import os
from dataclasses import dataclass, fields
import json
import sys

# Qt6 modules
from PySide6.QtWidgets import *

# Custom modules
from utils_qt_mk.config import _PACKAGE_DIR, cfd_data_file_path
_CFD_DATA_FILE = cfd_data_file_path()

from utils_qt_mk.general import SignalBlocker, Singleton, stub_repr, qt_connect


CFDType: _CFDType | None = None


def get_path_types(fetch_data: bool = False) -> list[str | CFDData]:
    """ Returns the available custom file dialog (CFD) types.

    :param fetch_data: A flag requesting the CFDData objects themselves.
        The default is False.
    """

    if fetch_data:
        return [pd for pd in PathTypes._path_types.values()]
    else:
        return [key.lower() for key in PathTypes._path_types.keys()]


def merge_json(path: str) -> None:
    """
    Takes an external JSON dialog data file and merges its contents to the
    package's own file. This way if you create dialogs you can reuse them in
    another project.

    :param path: Path to the external JSON dialog data file.
    """

    with open(_CFD_DATA_FILE, 'r') as f:
        package_data = json.load(f)

    package_data = [CFDData.from_dict(pdo) for pdo in package_data]

    with open(path, 'r') as f:
        external_data = json.load(f)

    external_data = [CFDData.from_dict(pdo) for pdo in external_data]

    pd_len = len(package_data)  # len used to not have to check newly...
    for pdo_e in external_data:  # ... appended items
        if all(not pdo_e.soft_eq(pdo_p) for pdo_p in package_data[:pd_len]):
            package_data.append(pdo_e)

    package_data = [pdo.as_dict for pdo in package_data]

    with open(_CFD_DATA_FILE, 'w') as f:
        json.dump(package_data, f, indent=4)

    try:
        os.remove(os.path.join(_PACKAGE_DIR, 'custom_file_dialog.pyi'))
    except FileNotFoundError:
        pass
    else:
        _init_module()  # Reinitialize the module so that types are reloaded


@dataclass
class CFDData:
    """ Data describing a configuration for a given custom file dialog (CFD).

    :param path_id: Unique text identifier of the path.
    :param window_title: Title to set for the dialog.
    :param dialog_type: Numeric identifier for open/save file or open directory.
    :param file_type_filter: String filter for file handler dialogs
        (e.g. "JSON (*.json)").
    :param path: Absolute path to start browsing from.
    """

    path_id: str
    window_title: str
    dialog_type: int
    file_type_filter: str = ''
    path: str = 'C:/'

    def __eq__(self, other) -> bool:
        """ Custom comparison rule, comparing each field. """

        if not isinstance(other, CFDData):
            return False

        return all(getattr(self, f.name) == getattr(other, f.name)
                   for f in fields(self))

    def soft_eq(self, other) -> bool:
        """ Custom soft comparison rule, comparing only the ID. """

        if not isinstance(other, CFDData):
            return False

        return self.path_id == other.path_id  # type: ignore

    @property
    def as_dict(self) -> dict:
        """ Returns a dictionary containing the set values. """

        return {f.name: getattr(self, f.name) for f in fields(self)}

    @classmethod
    def from_dict(cls, src: dict) -> CFDData:
        """ Returns an instance built from a dictionary.

        :param src: A dictionary containing data to build an instance,
            extracted from the handled JSON file.
        """

        return cls(**src)


def _import_json(full_id_key: bool = False) -> dict[str, CFDData] | None:
    """ Imports data from the handled JSON file.

    :param full_id_key: A flag marking whether to keep the full ID as key
        or to format it (default) beforehand.

    :returns: A dictionary with keys of path IDs and values of CFDData objects,
        imported from the handled JSON file (or None if there is no such file).
    """

    try:
        with open(_CFD_DATA_FILE, 'r') as f:
            data = json.load(f)
    except FileNotFoundError:
        return
    else:
        types_dict = {}
        for path_item in data:
            if full_id_key:
                key = path_item['path_id']
            else:
                path_type = path_item['path_id'][0]
                key = (path_item['path_id'].split('_', 1)[1]
                       .capitalize().replace('_', ' '))
                key = f"[{path_type}] {key}"

            types_dict[key] = CFDData(**path_item)

        return types_dict


class _FileDialogDataEditor(QDialog):
    """ An editor for developer use. """

    def __init__(self) -> None:
        """ Initializer for the class. """

        super().__init__(parent=None)

        self.setWindowTitle("File dialog data editor")
        self.setFixedWidth(400)

        self._file_dialog_types = _import_json()

        self._setup_ui()
        self._setup_connections()

    def _setup_ui(self) -> None:
        """ Sets up the user interface: GUI objects and layouts. """

        # GUI objects
        self._chkNewType = QCheckBox(text="New type", parent=None)
        self._cmbTypeList = QComboBox()
        self._cmbPathCategory = QComboBox()
        self._cmbPathCategory.addItems(['Source', 'Destination'])
        self._ledPathType = QLineEdit()
        self._ledPathType.setPlaceholderText("Path type")
        self._ledWindowTitle = QLineEdit()
        self._ledWindowTitle.setPlaceholderText("Window title")
        self._cmbDialogTypes = QComboBox()
        self._cmbDialogTypes.addItems(["Open file name",
                                       "Save file name",
                                       "Existing directory"])
        self._ledFileTypeFilter = QLineEdit()
        self._ledFileTypeFilter.setPlaceholderText("CSV (*.csv)")
        self._ledPath = QLineEdit()
        self._ledPath.setPlaceholderText('Path')
        self._btnDelete = QPushButton('Delete')
        self._btnExport = QPushButton('Export')

        # Layouts
        self._vloMainLayout = QVBoxLayout()
        self._vloMainLayout.addWidget(self._chkNewType)
        self._vloMainLayout.addWidget(self._cmbTypeList)
        self._vloMainLayout.addWidget(self._cmbPathCategory)
        self._vloMainLayout.addWidget(self._ledPathType)
        self._vloMainLayout.addWidget(self._ledWindowTitle)
        self._vloMainLayout.addWidget(self._cmbDialogTypes)
        self._vloMainLayout.addWidget(self._ledFileTypeFilter)
        self._vloMainLayout.addWidget(self._ledPath)
        self._vloMainLayout.addWidget(self._btnDelete)
        self._vloMainLayout.addWidget(self._btnExport)

        self.setLayout(self._vloMainLayout)

        # Further initialization
        if self._file_dialog_types is None:
            self._chkNewType.setChecked(True)
            self._cmbTypeList.setVisible(False)
        else:
            self._update_type_list_combobox()
            self._slot_type_selection_changed()

        self._slot_new_type_toggled()

    def _setup_connections(self) -> None:
        """ Sets up the connections of the GUI objects. """

        qt_connect(self._chkNewType.stateChanged, self._slot_new_type_toggled)
        qt_connect(self._cmbTypeList.currentIndexChanged,
                   self._slot_type_selection_changed)
        qt_connect(self._btnDelete.clicked, self._slot_delete_data)
        qt_connect(self._btnExport.clicked, self._slot_export_data)

    def _export_json(self) -> None:
        """ Exports data to the handled JSON file. """

        if self._file_dialog_types is None:
            os.remove(_CFD_DATA_FILE)
            return

        with open(_CFD_DATA_FILE, 'w') as f:
            json.dump([t.as_dict for t in self._file_dialog_types.values()],
                      f, indent=4)

    def _reset_inputs(self) -> None:
        """ Resets the input fields to their default values. """

        self._cmbPathCategory.setCurrentIndex(0)
        self._ledPathType.setText('')
        self._ledWindowTitle.setText('')
        self._cmbDialogTypes.setCurrentIndex(0)
        self._ledFileTypeFilter.setText('')
        self._ledPath.setText('')

    def _slot_new_type_toggled(self) -> None:
        """
        Sets the visibility of the type selector based on the control combobox.
        """

        self._cmbTypeList.setVisible(not self._chkNewType.isChecked())
        if not self._chkNewType.isChecked():  # Update by the currently...
            self._slot_type_selection_changed()  # ...selected existing theme

    def _slot_type_selection_changed(self) -> None:  # Index unused: not a param
        """ Updates the GUI according to the control combobox. """

        path_data: CFDData = self._file_dialog_types[
            self._cmbTypeList.currentText()]
        self._cmbPathCategory.setCurrentIndex(path_data.path_id.startswith('D'))
        self._ledPathType.setText(self._cmbTypeList.currentText()
                                  .split(' ', 1)[1])
        self._ledWindowTitle.setText(path_data.window_title)
        self._cmbDialogTypes.setCurrentIndex(path_data.dialog_type)
        self._ledFileTypeFilter.setText(path_data.file_type_filter)
        self._ledPath.setText(path_data.path)

    def _update_type_list_combobox(self) -> None:
        """ Updates the combobox from the type list. """

        with SignalBlocker(self._cmbTypeList) as obj:
            obj.clear()
            if self._file_dialog_types is not None:
                obj.addItems(self._file_dialog_types.keys())  # type: ignore
                obj.setCurrentIndex(obj.count() - 1)

    def _slot_delete_data(self) -> None:
        """ Attempts to delete the set data, updating the GUI. """

        pt = self._ledPathType.text()
        pt = f"[{self._cmbPathCategory.currentText()[0]}] {pt}"
        try:
            del self._file_dialog_types[pt]
        except (TypeError, KeyError):  # TypeError if None
            return
        else:
            if not self._file_dialog_types:  # Empty list
                self._file_dialog_types = None

            self._export_json()
            self._update_type_list_combobox()
            if self._file_dialog_types is not None:
                self._slot_type_selection_changed()
            else:
                self._chkNewType.setChecked(True)
                self._reset_inputs()

    def _slot_export_data(self) -> None:
        """
        Adds the set data to the stored dictionary and exports it, updating
        the GUI.
        """

        pc = self._cmbPathCategory.currentText().upper()
        pt = self._ledPathType.text()
        path_data = CFDData(f"{pc}_{pt.upper().replace(' ', '_')}",
                            self._ledWindowTitle.text(),
                            self._cmbDialogTypes.currentIndex(),
                            self._ledFileTypeFilter.text(),
                            self._ledPath.text())

        pt = f"[{self._cmbPathCategory.currentText()[0]}] {pt}"
        try:
            self._file_dialog_types[pt] = path_data
        except TypeError:  # If None
            self._file_dialog_types = {pt: path_data}

        self._export_json()
        self._update_type_list_combobox()
        self._chkNewType.setChecked(False)


class _CFDType(metaclass=Singleton):
    """ A class for Enum-like access to custom file dialog (CFD) types. """

    def __init__(self) -> None:
        """ Initializer for the class. """

        self._path_types = _import_json(full_id_key=True)

    def __getattr__(self, name: str) -> CFDData | None:
        """
        Returns the CFDData object identified by the passed string if there are
        CFD types loaded.

        :param name: The unique identifier of a path.
        """

        if self._path_types is not None:
            return self._path_types[name.upper()]


def custom_dialog(parent: QWidget, cfd_data: CFDData,
                  custom_title: str = None) -> tuple[bool, str | None]:
    """ Opens a file dialog of the requested type.

    :param parent: The widget from which the dialog is requested.
    :param cfd_data: An object defining the appearance and path of the dialog.
    :param custom_title: A custom title for the dialog. The default is None,
        which means that the one defined in the 'path_data' is used.

    :returns: A tuple containing the success flag and the selected path or None
        if the selection is unsuccessful.
    """

    selection_successful = False
    if custom_title is None:
        window_title = cfd_data.window_title
    else:
        window_title = custom_title

    path = ''
    if (dialog_type := cfd_data.dialog_type) == 0:  # Open file name
        path = QFileDialog.getOpenFileName(parent,  # type: ignore
                                           window_title,
                                           cfd_data.path,
                                           cfd_data.file_type_filter)
    elif dialog_type == 1:  # Save file name
        path = QFileDialog.getSaveFileName(parent,  # type: ignore
                                           window_title,
                                           cfd_data.path,
                                           cfd_data.file_type_filter)
    elif dialog_type == 2:  # Existing directory
        path = QFileDialog.getExistingDirectory(parent,  # type: ignore
                                                window_title,
                                                cfd_data.path)

    if dialog_type <= 1:
        selection_successful = path[0] != ''
        path = path[0]
    elif dialog_type == 2:
        selection_successful = path != ''

    if not selection_successful:
        return False, None

    new_path = ''
    if dialog_type <= 1:
        path_split = path.split('/')
        for i in range(len(path_split) - 1):
            new_path += path_split[i] + '/'
    elif dialog_type == 2:
        new_path = path

    with open(_CFD_DATA_FILE, 'r+') as f:
        data = json.load(f)
        for idx, entry in enumerate(data):
            if entry['path_id'] == cfd_data.path_id:
                data[idx]['path'] = new_path
                break

        f.seek(0)  # Jump to the beginning of the file
        json.dump(data, f, indent=4)

    return True, path


class _TestApplication(QMainWindow):
    """ The entry point for testing. """

    def __init__(self) -> None:
        """ Initializer for the class. """

        super().__init__(parent=None)

        self.setWindowTitle("Test application")

        # GUI and layouts
        self._setup_ui()
        self._setup_connections()

    def _setup_ui(self) -> None:
        """ Sets up the user interface: GUI objects and layouts. """

        # GUI objects
        self._btnDataEditor = QPushButton("Open a data editor dialog")

        # Layouts
        self._vloMainLayout = QVBoxLayout()
        self._vloMainLayout.addWidget(self._btnDataEditor)

        self._wdgCentralWidget = QWidget()
        self._wdgCentralWidget.setLayout(self._vloMainLayout)
        self.setCentralWidget(self._wdgCentralWidget)

    def _setup_connections(self) -> None:
        """ Sets up the connections of the GUI objects. """

        qt_connect(self._btnDataEditor.clicked, self._slot_de_test)

    @classmethod
    def _slot_de_test(cls) -> None:
        """ Tests the data editor dialog. """

        de = _FileDialogDataEditor()
        de.exec()


def _init_module():
    """ Initializes the module. """

    if not os.path.exists(os.path.join(_PACKAGE_DIR, 'custom_file_dialog.pyi')):
        imports = "from dataclasses import dataclass\n" \
                  "from PySide6.QtWidgets import QDialog, QMainWindow, " \
                  "QWidget\n" \
                  "from utils_qt_mk._general import Singleton\n\n\n"

        functions = [get_path_types, merge_json, _import_json, custom_dialog]
        reprs = [stub_repr(func) for func in functions]
        reprs.append('\n\n')

        classes = {CFDData: None,
                   _FileDialogDataEditor: None,
                   _CFDType: None,
                   _TestApplication: None}

        class_reprs = []
        for cls, sigs in classes.items():
            if cls == _CFDType:
                try:
                    with open(_CFD_DATA_FILE, 'r') as f:
                        data = json.load(f)
                except FileNotFoundError:
                    extra_cvs = None
                else:
                    extra_cvs = '\n'.join([f"\t{path_item['path_id'].lower()}: "
                                           "CFDData = None"
                                           for path_item in data])
            else:
                extra_cvs = None

            class_reprs.append(
                stub_repr(cls, signals=sigs, extra_cvs=extra_cvs))

        reprs.append('\n\n'.join(class_reprs))

        with open(os.path.join(_PACKAGE_DIR, 'custom_file_dialog.pyi'),
                  'w') as f:
            f.write(imports)
            f.write("CFDType: _CFDType = None\n\n\n")
            f.write(''.join(reprs))

    global PathTypes
    PathTypes = _CFDType()


_init_module()


if __name__ == '__main__':
    app = QApplication(sys.argv)  # type: ignore
    app.setStyle('Fusion')
    mainWindow = _TestApplication()
    mainWindow.show()
    app.exec()
